import tensorflow as tf
from kgcnn.layers.base import GraphBaseLayer
from kgcnn.layers.modules import Activation

ks = tf.keras


@ks.utils.register_keras_serializable(package='kgcnn', name='RelationalDense')
class RelationalDense(GraphBaseLayer):
    r"""Relational :obj:`Dense` layer for (ragged) tensors, representing embeddings or features.

    A :obj:`RelationalDense` layer computes a densely-connected NN layer, i.e. a linear transformation of the input
    :math:`\mathbf{x}` with the kernel weights matrix :math:`\mathbf{W}_r` and bias :math:`\mathbf{b}`
    plus (possibly non-linear) activation function :math:`\sigma` for each type of relation :math:`r` that underlies
    the feature or embedding. Examples are different edge or node types such as chemical bonds and atomic species.

    .. math::

        \mathbf{x}'_r = \sigma (\mathbf{x}_r \mathbf{W}_r + \mathbf{b})

    This has been proposed by `Schlichtkrull et al. (2017) <https://arxiv.org/abs/1703.06103>`_ for graph networks.
    Additionally, there are a set of regularization schemes to improve performance and reduce the number of learnable
    parameters proposed by `Schlichtkrull et al. (2017) <https://arxiv.org/abs/1703.06103>`_ .
    Here, the following is implemented: basis-, block-diagonal-decomposition.
    With the basis decom-position, each :math:`\mathbf{W}_r` is defined as follows:

     .. math::

        \mathbf{W}_r = \sum_{b=1}^{B} a_{rb}\; \mathbf{V}_b

    i.e. as a linear combination of basis transformations :math:`V_b \in \mathbb{R}^{d' \times d}` with coefficients
    :math:`a_{rb}` such that only the coefficients depend on :math:`r`.
    In the block-diagonal decomposition, let each :math:`W_r` be defined through the direct sum over a set of
    low-dimensional matrices:

    .. math::

        \mathbf{W}_r = \otimes_{b=1}^{B} \mathbf{Q}_{br}

    Thereby, :math:`W_r` are block-diagonal matrices: :math:`diag(Q_{1r} , \dots , Q_{Br})` with
    :math:`Q_{br} \in \mathbb{R}^{(d'/B)\times(d/B)}`.
    Usage:

    .. code-block:: python

        import tensorflow as tf
        from kgcnn.layers.relational import RelationalDense
        f = tf.ragged.constant([[[1.0, 0.0 ]], [[1.0, 4.0], [2.0, 2.0]], [[3.0, 5.0]]],
            inner_shape=(2,), ragged_rank=1)
        r = tf.ragged.constant([[1], [1, 2], [3]], ragged_rank=1)
        out = RelationalDense(6, num_relations=5, num_bases=3, num_blocks=2)([f, r])
    """

    def __init__(self,
                 units: int,
                 num_relations: int,
                 num_bases: int = None,
                 num_blocks: int = None,
                 activation=None,
                 use_bias: bool = True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        r"""Initialize layer similar to :obj:`ks.layers.Dense`.

         Args:
            units: Positive integer, dimensionality of the output space.
            num_relations: Number of relations expected to construct weights.
            num_bases: Number of kernel basis functions to construct relations. Default is None.
            num_blocks:  Number of block-matrices to get
            activation: Activation function to use.
                If you don't specify anything, no activation is applied
                (ie. "linear" activation: `a(x) = x`).
            use_bias: Boolean, whether the layer uses a bias vector.
            kernel_initializer: Initializer for the `kernel` weights matrix.
            bias_initializer: Initializer for the bias vector.
            kernel_regularizer: Regularizer function applied to
                the `kernel` weights matrix.
            bias_regularizer: Regularizer function applied to the bias vector.
            activity_regularizer: Regularizer function applied to
                the output of the layer (its "activation").
            kernel_constraint: Constraint function applied to
                the `kernel` weights matrix.
            bias_constraint: Constraint function applied to the bias vector.
        """
        super(RelationalDense, self).__init__(**kwargs)
        self.units = units
        self.use_bias = use_bias
        self.num_relations = num_relations
        self.num_bases = num_bases
        self.num_blocks = num_blocks
        self.kernel_initializer = ks.initializers.get(kernel_initializer)
        self.bias_initializer = ks.initializers.get(bias_initializer)
        self.kernel_regularizer = ks.regularizers.get(kernel_regularizer)
        self.bias_regularizer = ks.regularizers.get(bias_regularizer)
        self.kernel_constraint = ks.constraints.get(kernel_constraint)
        self.bias_constraint = ks.constraints.get(bias_constraint)
        self._layer_activation = Activation(activation=activation, activity_regularizer=activity_regularizer)

    def build(self, input_shape):
        """Build layer, i.e. check input and construct weights for this layer.

        Args:
            input_shape: Shape of the input.

        Returns:
            None.
        """
        assert len(input_shape) == 2, "`RelationalDense` layer requires feature plus relation information."
        dtype = tf.as_dtype(self.dtype or ks.backend.floatx())
        if not (dtype.is_floating or dtype.is_complex):
            raise TypeError(
                "A Dense layer can only be built with a floating-point "
                f"dtype. Received: dtype={dtype}"
            )

        feature_shape = tf.TensorShape(input_shape[0])
        relation_shape = tf.TensorShape(input_shape[1])
        last_dim = tf.compat.dimension_value(feature_shape[-1])
        if last_dim is None:
            raise ValueError(
                "The last dimension of the inputs to a Dense layer "
                "should be defined. Found None. "
                f"Full input shape received: {feature_shape}"
            )
        assert feature_shape.rank - 1 == relation_shape.rank, "Relations must be without feature dimension."

        # Find kernel shape.
        if self.num_bases is not None:
            num_multi_kernels = self.num_bases
        else:
            num_multi_kernels = self.num_relations
        if self.num_blocks is not None:
            assert (last_dim % self.num_blocks == 0
                    and self.units % self.num_blocks == 0), "`num_blocks` must divide in- and output dimension."
            in_kernel_dim = int(last_dim / self.num_blocks)
            out_kernel_dim = int(self.units / self.num_blocks)
        else:
            in_kernel_dim = last_dim
            out_kernel_dim = self.units
        kernel_shape = [num_multi_kernels, in_kernel_dim, out_kernel_dim]
        if self.num_blocks is not None:
            kernel_shape = [num_multi_kernels, self.num_blocks, in_kernel_dim, out_kernel_dim]
        # Make kernel
        self.kernel = self.add_weight(
            "kernel",
            shape=kernel_shape,
            initializer=self._multi_kernel_initializer,
            regularizer=self.kernel_regularizer,
            constraint=self.kernel_constraint,
            dtype=self.dtype, trainable=True,
        )
        if self.num_bases is not None:
                self.lin_bases = self.add_weight(
                "lin_bases",
                shape=[self.num_relations, self.num_bases],
                initializer=self.kernel_initializer,
                regularizer=self.kernel_regularizer,
                constraint=self.kernel_constraint,
                dtype=self.dtype, trainable=True,
            )
        else:
            self.lin_bases = None
        if self.use_bias:
            self.bias = self.add_weight(
                "bias",
                shape=[self.units, ],
                initializer=self.bias_initializer,
                regularizer=self.bias_regularizer,
                constraint=self.bias_constraint,
                dtype=self.dtype, trainable=True,
            )
        else:
            self.bias = None

        self.built = True

    def _multi_kernel_initializer(self, shape, dtype=None, **kwargs):
        """Initialize multiple relational kernels.

        Args:
            shape: Shape of multi-kernel tensor.
            dtype: Optional dtype of the tensor.
            kwargs: Additional keyword arguments.

        Returns:
            tf.Tensor: Tensor for initialization.
        """
        # For blocks the ks.initialize seems to have proper behaviour.
        # Each block is treated as a convolution field.
        if len(shape) < 3:
            return self.kernel_initializer(shape, dtype=dtype, **kwargs)
        # Initialize each kernel separately.
        separate_kernels = [
            tf.expand_dims(self.kernel_initializer(shape[1:], dtype=dtype, **kwargs), axis=0) for _ in range(shape[0])]
        relational_kernel = tf.concat(separate_kernels, axis=0)
        return relational_kernel

    def _construct_kernel_weights(self):
        """Assemble the actual relational kernel from bases or blocks."""
        if self.num_blocks is not None:
            matrix_list = [
                tf.linalg.LinearOperatorFullMatrix(
                    tf.gather(self.kernel, i, axis=1)) for i in range(self.kernel.shape[1])
            ]
            operator = tf.linalg.LinearOperatorBlockDiag(matrix_list)
            kernel = operator.to_dense()
        else:
            kernel = self.kernel
        if self.num_bases is not None:
            kernel = tf.tensordot(self.lin_bases, kernel, [1, 0])
        return kernel

    def call(self, inputs, **kwargs):
        r"""Forward pass. Here, the relation is assumed to be encoded at axis=1.

        Args:
            inputs: [features, relations]

                - features (tf.RaggedTensor, tf.Tensor): Feature tensor of shape `(batch, [N], F)` of dtype 'float'.
                - relations (tf.RaggedTensor, tf.Tensor): Relation tensor of shape `(batch, [N])` of dtype 'int'.

        Returns:
            tf.RaggedTensor: Processed feature tensor. Shape is `(batch, [N], units)` of dtype 'float'.
        """
        features, relations = self.assert_ragged_input_rank(inputs, ragged_rank=1)
        built_kernel = self._construct_kernel_weights()
        kernel_per_feature = tf.gather(built_kernel, relations.values)
        new_features = ks.backend.batch_dot(features.values, kernel_per_feature)
        if self.use_bias:
            new_features = new_features + self.bias
        new_features = self._layer_activation(new_features, **kwargs)
        return tf.RaggedTensor.from_row_splits(new_features, features.row_splits, validate=self.ragged_validate)

    def get_config(self):
        config = super().get_config()
        config.update({
            "units": self.units,
            "use_bias": self.use_bias,
            "num_relations": self.num_relations,
            "num_bases": self.num_bases,
            "num_blocks": self.num_blocks,
            "kernel_initializer": ks.initializers.serialize(self.kernel_initializer),
            "bias_initializer": ks.initializers.serialize(self.bias_initializer),
            "kernel_regularizer": ks.regularizers.serialize(self.kernel_regularizer),
            "bias_regularizer": ks.regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": ks.constraints.serialize(self.kernel_constraint),
            "bias_constraint": ks.constraints.serialize(self.bias_constraint),
        })
        config_act = self._layer_activation.get_config()
        config.update({
            "activation": config_act["activation"],
            "activity_regularizer": config_act["activity_regularizer"]
        })
        return config



import unittest

import numpy as np
import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.xai.testing import MockContext


class TestMockContext(unittest.TestCase):

    def test_basically_works(self):
        num_elements = 10
        num_targets = 2
        with MockContext(num_elements=num_elements, num_targets=num_targets) as mock:
            assert isinstance(mock, MockContext)

            assert isinstance(mock.model, ks.models.Model)
            assert mock.model.built

            assert isinstance(mock.x, tuple)
            assert len(mock.x) == 3
            for element in mock.x:
                print(element.shape)
                assert isinstance(element, tf.RaggedTensor)

            assert isinstance(mock.y, np.ndarray)

            targets = mock.y
            assert isinstance(targets, np.ndarray)
            assert len(targets.shape) == 2
            assert targets.shape[-1] == num_targets



import random
import typing as t

import numpy as np
import tensorflow as tf
import tensorflow.keras as ks

from kgcnn.layers.conv.gat_conv import AttentionHeadGATV2
from kgcnn.layers.modules import DenseEmbedding
from kgcnn.layers.pooling import PoolingGlobalEdges
from kgcnn.data.utils import ragged_tensor_from_nested_numpy


# This is a very simple mock implementation, because to test the explanation methods we need some sort
# of a model as basis and this model will act as such.
class Model(ks.models.Model):

    def __init__(self,
                 num_targets: int = 1):
        super(Model, self).__init__()
        self.conv_layers = [
            AttentionHeadGATV2(units=64, use_edge_features=True, use_bias=True),
        ]
        self.lay_pooling = PoolingGlobalEdges(pooling_method='sum')
        self.lay_dense = DenseEmbedding(units=num_targets, activation='linear')

    def call(self, inputs, training=False):
        node_input, edge_input, edge_index_input = inputs
        x = node_input
        for lay in self.conv_layers:
            x = lay([x, edge_input, edge_index_input])

        pooled = self.lay_pooling(x)
        out = self.lay_dense(pooled)
        return out


class MockContext:

    def __init__(self,
                 num_elements: int = 10,
                 num_targets: int = 1,
                 epochs: int = 10,
                 batch_size: int = 2):
        self.num_elements = num_elements
        self.num_targets = num_targets
        self.epochs = epochs
        self.batch_size = batch_size

        self.model = Model(num_targets=num_targets)
        self.x = None
        self.y = None

    def generate_graph(self,
                       num_nodes: int,
                       num_node_attributes: int = 3,
                       num_edge_attributes: int = 1):
        remaining = list(range(num_nodes))
        random.shuffle(remaining)
        inserted = [remaining.pop(0)]
        node_attributes = [[random.random() for _ in range(num_node_attributes)] for _ in range(num_nodes)]
        edge_indices = []
        edge_attributes = []
        while len(remaining) != 0:
            i = remaining.pop(0)
            j = random.choice(inserted)
            inserted.append(i)

            edge_indices += [[i, j], [j, i]]
            edge_attribute = [1 for _ in range(num_edge_attributes)]
            edge_attributes += [edge_attribute, edge_attribute]

        return (
            np.array(node_attributes, dtype=float),
            np.array(edge_attributes, dtype=float),
            np.array(edge_indices, dtype=int)
        )

    def generate_data(self):
        node_attributes_list = []
        edge_attributes_list = []
        edge_indices_list = []
        targets_list = []
        for i in range(self.num_elements):
            num_nodes = random.randint(5, 20)
            node_attributes, edge_attributes, edge_indices = self.generate_graph(num_nodes)
            node_attributes_list.append(node_attributes)
            edge_attributes_list.append(edge_attributes)
            edge_indices_list.append(edge_indices)

            # The target value we will actually determine deterministically here so that our network
            # actually has a chance to learn anything
            target = np.sum(node_attributes)
            targets = [target for _ in range(self.num_targets)]
            targets_list.append(targets)

        self.x = (
            ragged_tensor_from_nested_numpy(node_attributes_list),
            ragged_tensor_from_nested_numpy(edge_attributes_list),
            ragged_tensor_from_nested_numpy(edge_indices_list)
        )

        self.y = (
            np.array(targets_list, dtype=float)
        )

    def __enter__(self):
        # This method will generate random input and output data and thus populate the internal attributes
        # self.x and self.y
        self.generate_data()

        # Using these we will train our mock model for a few very brief epochs.
        self.model.compile(
            loss=ks.losses.mean_squared_error,
            metrics=ks.metrics.mean_squared_error,
            run_eagerly=False,
            optimizer=ks.optimizers.Nadam(learning_rate=0.01),
        )
        hist = self.model.fit(
            self.x, self.y,
            batch_size=self.batch_size,
            epochs=self.epochs,
            verbose=0,
        )
        self.history = hist.history

        return self

    def __exit__(self, *args, **kwargs):
        pass



from kgcnn.data.datasets.MoleculeNetDataset2018 import MoleculeNetDataset2018


class ESOLDataset(MoleculeNetDataset2018):
    r"""Store and process 'ESOL' dataset from `MoleculeNet <https://moleculenet.org/>`_ database.
    Class inherits from :obj:`MoleculeNetDataset2018` and downloads dataset on class initialization.

    Compare reference:
    `DeepChem <https://deepchem.readthedocs.io/en/latest/api_reference/moleculenet.html#delaney-datasets>`__
    reading:

    Water solubility data(log solubility in mols per litre) for common organic small molecules.
    Random or Scaffold splitting is recommended for this dataset.
    Description in DeepChem reads: 'The Delaney (ESOL) dataset a regression dataset containing structures and water
    solubility data for 1128 compounds. The dataset is widely used to validate machine learning models on
    estimating solubility directly from molecular structures (as encoded in SMILES strings).'

    References:

        (1) Delaney, John S. ESOL: estimating aqueous solubility directly from molecular structure.
            Journal of chemical information and computer sciences 44.3 (2004): 1000-1005.

    """

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize ESOL dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        super(ESOLDataset, self).__init__("ESOL", reload=reload, verbose=verbose)



"""
"Ying et al. - GNNExplainer: Generating Explanations for Graph Neural Networks"

**Changelog**

??.??.2022 - Initial implementation

30.01.2023 - Added the class "GnnExplainer" which supports RaggedTensors and can thus generate multiple
explanations at once, greatly improving time efficiency for explaining large batches of predictions. However
the new class does not implement visualization of the explanations. This will have to be realized on a
higher abstraction level.
"""
import time
import typing as t

import numpy as np
import tensorflow as tf
ks = tf.keras

from kgcnn.xai.base import ImportanceExplanationMethod

# Keep track of model version from commit date in literature.
# To be updated if model is changed in a significant way.
__model_version__ = "2023.01.30"


# == REDUCED, RAGGED TENSOR IMPLEMENTATION ==

class GnnExplainer(ImportanceExplanationMethod):
    """
    Implementation of "ImportanceExplanationMethod", which means that calling an instance of this class
    given a model, a ragged input tensor and output predictions, it should return the corresponding
    node and edge importance tensors, which provide an explanation by assigning each node and edge of the
    input graphs with a 0-1 importance value.

    By the nature of the base idea behind GNNExplainer, the number of explanations produced has to be equal
    to the number of prediction targets that are generated by the model. Each target will receive its own
    explanation.
    """
    def __init__(self,
                 channels: int,
                 epochs: int = 100,
                 learning_rate: float = 0.01,
                 node_sparsity_factor: float = 0.1,
                 edge_sparsity_factor: float = 0.1,
                 log_step: int = 10,
                 verbose: bool = True):
        super(GnnExplainer, self).__init__(channels=channels)
        self.epochs = epochs
        self.learning_rate = learning_rate
        self.log_step = log_step
        self.verbose = verbose
        self.node_sparsity_factor = node_sparsity_factor
        self.edge_sparsity_factor = edge_sparsity_factor

    def __call__(self,
                 model: ks.models.Model,
                 x: t.Tuple[tf.RaggedTensor, tf.RaggedTensor, tf.RaggedTensor],
                 y: np.ndarray):
        """Given a model, the input tensor and the output array, this method will return a tuple of two
        ragged tensors, which represent the node importances and the edge importances.

        Beware, that this method executes an entire training process and may take some time.

        Reference of tensor shapes. [Brackets] indicate ragged dimension

            - V: Number of nodes in graph
            - E: Number of edges in graph
            - K: Number of explanation channels given in constructor. This has to be equal to the number of
                prediction targets specified in the constructor.
            - N: Number of node attributes
            - M: Number of edge attributes
            - B: batch size

        Args:
            x: A tuple (node_input, edge_input, edge_indices) of 3 RaggedTensors

                - node_input: Shape ([B], [V], N)
                - edge_input: Shape ([B], [E], M)
                - edge_indices: Shape ([B], [E], 2)

            y: A numpy array of shape (B, K)
            model: Any compatible keras model, which means any model which accepts the previously described
                input tensors and returns output similar to the previously described output tensor.

        Returns:
            A tuple (node_importances, edge_importances) of RaggedTensors.

                - node_importances: Shape ([B], [V], K)
                - edge_importances: Shape ([B], [E], K)

        """
        # Generally the idea of the implementation is that we use the node_input and edge_input tensors as
        # templates to generate the mask variable tensors, which match the graph dimensions but differ in
        # the final dimension, which instead of the node / edge features we will use to represent the
        # number of importance channels (== number of prediction targets).

        node_input, edge_input, edge_indices = x

        # Here we reduce away the last dimension of node and edge input to get just the ragged graph sizes
        # But we run into a problem here with multiple channels: We cant actually use the last dimension to
        # represent the number of different explanation channels. Instead, we do a workaround here where for
        # each channel we extend the batch dimension. Aka we assume that all the different channels are just
        # additional graphs to be treated like the others. The reason why we have to do it like that is
        # because later on we need to multiply the masks with the inputs!
        node_mask_single = tf.reduce_mean(tf.ones_like(node_input), axis=-1, keepdims=True)
        node_mask_ragged = tf.concat([node_mask_single for _ in range(self.channels)], axis=0)
        node_mask_variables = tf.Variable(node_mask_ragged.flat_values, trainable=True, dtype=tf.float64)

        edge_mask_single = tf.reduce_mean(tf.ones_like(edge_input), axis=-1, keepdims=True)
        edge_mask_ragged = tf.concat([edge_mask_single for _ in range(self.channels)], axis=0)
        edge_mask_variables = tf.Variable(edge_mask_ragged.flat_values, trainable=True, dtype=tf.float64)

        optimizer = ks.optimizers.Nadam(learning_rate=self.learning_rate)

        # This is a logical extension of what was previously described. Since we treat the different
        # explanation channels as just a batch extension, we have to modify the input values and the output
        # values accordingly so that they have the same batch size so to say. Naturally we simply have to
        # duplicate the values.
        x_extended = (
            tf.concat([node_input for _ in range(self.channels)], axis=0),
            tf.concat([edge_input for _ in range(self.channels)], axis=0),
            tf.concat([edge_indices for _ in range(self.channels)], axis=0),
        )
        y_extended = []
        for c in range(self.channels):
            y_mod = np.zeros_like(y)
            y_mod[:, c] = y[:, c]
            y_extended.append(y_mod)

        y_extended = np.concatenate(y_extended)

        start_time = time.time()
        for epoch in range(self.epochs):

            with tf.GradientTape() as tape:
                node_mask = tf.RaggedTensor.from_nested_row_splits(
                    node_mask_variables,
                    nested_row_splits=node_mask_ragged.nested_row_splits
                )

                edge_mask = tf.RaggedTensor.from_nested_row_splits(
                    edge_mask_variables,
                    nested_row_splits=edge_mask_ragged.nested_row_splits
                )

                out = model([
                    x_extended[0] * node_mask,
                    x_extended[1] * edge_mask,
                    x_extended[2]
                ])

                # The loss can basically be summerized as: We try to find the smallest subset of nodes and
                # edges in the input, which will cause the network to get as close as possible to it's
                # original prediction!
                loss = tf.cast(tf.reduce_mean(tf.square(y_extended - out)), dtype=tf.float64)
                # Important detail: The reduce_sum here reduces over all the nodes / edges and is necessary!
                loss += self.node_sparsity_factor * tf.reduce_mean(tf.reduce_sum(tf.abs(node_mask), axis=1))
                loss += self.edge_sparsity_factor * tf.reduce_mean(tf.reduce_sum(tf.abs(edge_mask), axis=1))

            trainable_vars = [node_mask_variables, edge_mask_variables]
            gradients = tape.gradient(loss, trainable_vars)
            optimizer.apply_gradients(zip(gradients, trainable_vars))

            if self.verbose and epoch % self.log_step == 0:
                print(f' * epoch ({epoch}/{self.epochs}) '
                      f' - loss: {loss}'
                      f' - elapsed time: {time.time()-start_time:.2f} seconds')

        # For the training we had to treat the different explanation channels as a batch extension. As per
        # the interface we need to return the importances however such that the different explanation
        # channels are organized into the third dimension of the tensors.

        # Sadly this does not work in a more direct fashion. We get the number of elements of nodes and
        # edges that belong to one explanation channel. Iterate in chunks of that size and turn each of
        # those chunks into it's own explanation respectively. At the end we concatenate all of them in
        # the 3rd dimension to produce the desired result.
        num_elements_node = node_mask_single.flat_values.shape[0]
        num_elements_edge = edge_mask_single.flat_values.shape[0]
        node_importances_list = []
        edge_importances_list = []
        for c in range(self.channels):
            node_importances_part = tf.RaggedTensor.from_nested_row_splits(
                node_mask_variables[c*num_elements_node:(c+1)*num_elements_node, :],
                node_mask_single.nested_row_splits
            )
            node_importances_list.append(node_importances_part)

            edge_importances_part = tf.RaggedTensor.from_nested_row_splits(
                edge_mask_variables[c*num_elements_edge:(c+1)*num_elements_edge, :],
                edge_mask_single.nested_row_splits
            )
            edge_importances_list.append(edge_importances_part)

        return (
            tf.concat(node_importances_list, axis=-1),
            tf.concat(edge_importances_list, axis=-1)
        )


# == ORIGINAL IMPLEMENTATION ==

class GNNInterface:
    """An interface class which should be implemented by a Graph Neural Network (GNN) model to make it explainable.
    This class is just an interface, which is used by the `GNNExplainer` and should be implemented in a subclass.
    The implementation of this class could be a wrapper around an existing Tensorflow/Keras GNN.
    The output of the methods `predict` and `masked_predict` should be of same dimension and the output to be explained.

    """

    def predict(self, gnn_input, **kwargs):
        """Returns the prediction for the `gnn_input`.

        Args:
            gnn_input: The input graph to which a prediction should be made by the GNN.

        Raises:
            NotImplementedError: This is just an interface class, to indicate which methods should be implemented.
                Implement this method in a subclass.
        """
        raise NotImplementedError(
            "Implement this method in a specific subclass.")

    def masked_predict(self, gnn_input, edge_mask, feature_mask, node_mask, **kwargs):
        """Returns the prediction for the `gnn_input` when it is masked by the three given masks.

        Args:
            gnn_input: The input graph to which should be masked before a prediction should be made by the GNN.
            edge_mask: A `Tensor` of shape `[get_number_of_edges(self, gnn_input), 1]`,
                which should mask the edges of the input graph.
            feature_mask: A `Tensor` of shape `[get_number_of_node_features(self, gnn_input), 1]`,
                which should mask the node features in the input graph.
            node_mask: A `Tensor` of shape `[get_number_of_nodes(self, gnn_input), 1]`,
                which should mask the node features in the input graph.

        Raises:
            NotImplementedError: This is just an interface class, to indicate which methods should be implemented.
                Implement this method in a subclass.
        """
        raise NotImplementedError(
            "Implement this method in a specific subclass.")

    def get_number_of_nodes(self, gnn_input):
        """Returns the number of nodes in the `gnn_input` graph.

        Args:
            gnn_input: The input graph to which this function returns the number of nodes in.

        Raises:
            NotImplementedError: This is just an interface class, to indicate which methods should be implemented.
                Implement this method in a subclass.
        """
        raise NotImplementedError(
            "Implement this method in a specific subclass.")

    def get_number_of_edges(self, gnn_input):
        """Returns the number of edges in the `gnn_input` graph.

        Args:
            gnn_input: The input graph to which this function returns the number of edges in.

        Raises:
            NotImplementedError: This is just an interface class, to indicate which methods should be implemented.
                Implement this method in a subclass.
        """
        raise NotImplementedError(
            "Implement this method in a specific subclass.")

    def get_number_of_node_features(self, gnn_input):
        """Returns the number of node features to the corresponding `gnn_input`.

        Args:
            gnn_input: The input graph to which this function returns the number of node features in.

        Raises:
            NotImplementedError: This is just an interface class, to indicate which methods should be implemented.
                Implement this method in a subclass.
        """
        raise NotImplementedError(
            "Implement this method in a specific subclass.")

    def get_explanation(self, gnn_input, edge_mask, feature_mask, node_mask, **kwargs):
        """Takes the graph input and the masks learned by the GNNExplainer and combines them to some sort of explanation
        The form of explanation could e.g. consist of a networkx graph,
        which has mask values as labels to nodes/edge and a dict for the feature explanation values.

        Args:
            gnn_input: The input graph to which should the masks were found by the GNNExplainer.
            edge_mask: A `Tensor` of shape `[get_number_of_edges(self, gnn_input), 1]`,
                which was found by the GNNExplainer.
            feature_mask: A `Tensor` of shape `[get_number_of_node_features(self, gnn_input), 1]`,
                which was found by the GNNExplainer.
            node_mask: A `Tensor` of shape `[get_number_of_nodes(self, gnn_input), 1]`,
                which was found by the GNNExplainer.

        Raises:
            NotImplementedError: This is just an interface class, to indicate which methods should be implemented.
                Implement this method in a subclass.
        """
        raise NotImplementedError(
            "Implement this method in a specific subclass.")

    def present_explanation(self, explanation, **kwargs):
        """Takes an explanation, which was generated by `get_explanation` and presents it to the user in a suitable way.
        The presentation of an explanation largely depends on the data domain and targeted user group.
        Examples for presentations:

        * A visualization of the most relevant subgraph(s) to the decision
        * A visualization of the whole graph with highlighted parts
        * Bar diagrams for feature explanations
        * ...

        Args:
            explanation: An explanation for the GNN decision,
                which is of the form the `get_explanation` method returns an explanation.
        Raises:
            NotImplementedError: This is just an interface class, to indicate which methods should be implemented.
                Implement this method in a subclass.
        """
        raise NotImplementedError(
            "Implement this method in a specific subclass.")


class GNNExplainer:
    """`GNNExplainer` explains the decisions of a GNN, which implements `GNNInterface`.

    See Ying et al. (https://arxiv.org/abs/1903.03894) for details on how such an explanation is found.
    Note that this implementation is inspired by the paper by Ying et al., but differs in some aspects.
    """

    def __init__(self, gnn, gnnexplaineroptimizer_options=None,
                 compile_options=None, fit_options=None, **kwargs):
        """Constructs a GNNExplainer instance for the given `gnn`.

        Args:
            gnn: An instance of a class which implements the `GNNInterface`.
            gnnexplaineroptimizer_options (dict, optional): Parameters in this dict are forwarded to the constructor
                of the `GNNExplainerOptimizer` (see docstring of `GNNExplainerOptimizer.__init__`).
                Defaults to {}.
            compile_options (dict, optional): Parameters in ths dict are forwarded to the `keras.Model.compile` method
                of the `GNNExplainerOpimizer`. Can be used to customize the optimization process of the
                `GNNExplainerOptimizer`.
                Defaults to {}.
            fit_options (dict, optional): Parameters in ths dict are forwarded to the `keras.Model.fit` method
                of the `GNNExplainerOpimizer`.
                Defaults to {}.
        """
        if gnnexplaineroptimizer_options is None:
            gnnexplaineroptimizer_options = {}
        if compile_options is None:
            compile_options = {}
        if fit_options is None:
            fit_options = {}
        self.gnn = gnn
        self.gnnx_optimizer = None
        self.graph_instance = None
        self.gnnexplaineroptimizer_options = gnnexplaineroptimizer_options
        self.compile_options = compile_options
        self.fit_options = fit_options

    def explain(self, graph_instance, output_to_explain=None, inspection=False, **kwargs):
        """Finds the masks to the decision of the `self.gnn` on the given `graph_instance`.
        This method does not have a return value, but only has side effects.
        To get the explanation which was found, call `get_explanation` after calling this method.
        This method just instantiates a `GNNExplainerOptimizer`,
        which then finds the masks for the explanation via gradient descent.

        Args:
            graph_instance: The graph input to the GNN to which an explanation should be found.
            output_to_explain (optional): Set this parameter to the output which should be explained.
                By default the GNNExplainer explains the output the `self.gnn` on the given `graph_instance`.
                Defaults to None.
            inspection (optional): If `inspection` is set to True this function will return information
                about the optimization process in a dictionary form.
                Be aware that inspections results in longer runtimes.
                Defaults to False.
        """

        self.graph_instance = graph_instance

        # Add inspection callback to fit options, if inspection is True
        fit_options = self.fit_options.copy()
        if inspection:
            inspection_callback = self.InspectionCallback(self.graph_instance)
            if 'callbacks' in self.fit_options.keys():
                fit_options['callbacks'].append(inspection_callback)
            else:
                fit_options['callbacks'] = [inspection_callback]

        # Set up GNNExplainerOptimizer and optimize with respect to masks
        gnnx_optimizer = GNNExplainerOptimizer(
            self.gnn, graph_instance, **self.gnnexplaineroptimizer_options)
        self.gnnx_optimizer = gnnx_optimizer
        if output_to_explain is not None:
            gnnx_optimizer.output_to_explain = output_to_explain
        gnnx_optimizer.compile(**self.compile_options)
        gnnx_optimizer.fit(graph_instance, **fit_options)

        # Read out information from inspection_callback
        if inspection:
            dict_fields = ['predictions',
                           'total_loss',
                           'edge_mask_loss',
                           'feature_mask_loss',
                           'node_mask_loss']
            inspection_information = {}
            for field in dict_fields:
                if hasattr(inspection_callback, field) and len(getattr(inspection_callback, field)) > 0:
                    inspection_information[field] = getattr(inspection_callback, field)
            return inspection_information

    def get_explanation(self, **kwargs):
        """Returns the explanation (derived from the learned masks) to a decision on the graph,
        which was passed to the `explain` method before.
        Important: The `explain` method should always be called before calling this method.
        Internally this method just calls the `GNNInterface.get_explanation` method
        implemented by the `self.gnn` with the masks found by the `GNNExplainerOptimizer` as parameters.

        Raises:
            Exception: If the `explain` method is not called before, this method raises an Exception.
        Returns:
            The explanation which is returned by `GNNInterface.get_explanation` implemented by the `self.gnn`,
            parametrized by the learned masks.
        """
        if self.graph_instance is None or self.gnnx_optimizer is None:
            raise Exception(
                "You must first call explain on the GNNExplainer instance.")

        edge_mask = self.gnnx_optimizer.get_mask("edge")
        feature_mask = self.gnnx_optimizer.get_mask("feature")
        node_mask = self.gnnx_optimizer.get_mask("node")

        return self.gnn.get_explanation(self.graph_instance,
                                        edge_mask,
                                        feature_mask,
                                        node_mask, **kwargs)

    def present_explanation(self, explanation, **kwargs):
        """Takes an explanation, which was generated by `get_explanation` and presents it.
        Internally this method just calls the `GNNInterface.present_explanation` method
        implemented by the `self.gnn`.

        Args:
            explanation: The explanation (obtained by `get_explanation`) which should be presented.
        Returns:
            A presentation of the given explanation.
        """
        return self.gnn.present_explanation(explanation, **kwargs)

    class InspectionCallback(ks.callbacks.Callback):
        """Callback class to get the inspection information,
        if 'inspection' is set to true for the 'GNNExplainer.explain' method.
        """

        def __init__(self, graph_instance):
            super(GNNExplainer.InspectionCallback, self).__init__()
            self.graph_instance = graph_instance
            self.predictions = []
            self.total_loss = []
            self.edge_mask_loss = []
            self.feature_mask_loss = []
            self.node_mask_loss = []

        def on_epoch_begin(self, epoch, logs=None):
            masked = self.model.call(self.graph_instance).numpy()[0]
            self.predictions.append(masked)

        def on_epoch_end(self, epoch, logs=None):
            """After epoch."""
            index = 0
            losses_list = [loss_iter.numpy() for loss_iter in self.model.losses]
            if self.model.edge_mask_loss_weight > 0:
                self.edge_mask_loss.append(losses_list[index])
                index = index + 1
            if self.model.feature_mask_loss_weight > 0:
                self.feature_mask_loss.append(losses_list[index])
                index = index + 1
            if self.model.node_mask_loss_weight > 0:
                self.node_mask_loss.append(losses_list[index])
            self.total_loss.append(logs['loss'])


class GNNExplainerOptimizer(ks.Model):
    """The `GNNExplainerOptimizer` solves the optimization problem which is used to find masks,
    which then can be used to explain decisions by GNNs.
    """

    def __init__(self, gnn_model, graph_instance,
                 edge_mask_loss_weight=1e-4,
                 edge_mask_norm_ord=1,
                 feature_mask_loss_weight=1e-4,
                 feature_mask_norm_ord=1,
                 node_mask_loss_weight=0.0,
                 node_mask_norm_ord=1,
                 **kwargs):
        """Constructs a `GNNExplainerOptimizer` instance with the given parameters.

        Args:
            gnn_model (GNNInterface): An instance of a class which implements the methods of the `GNNInterface`.
            graph_instance: The graph to which the masks should be found.
            edge_mask_loss_weight (float, optional): The weight of the edge mask loss term in the optimization problem.
                Defaults to 1e-4.
            edge_mask_norm_ord (float, optional): The norm p value for the p-norm, which is used on the edge mask.
                Smaller values encourage sparser masks.
                Defaults to 1.
            feature_mask_loss_weight (float, optional): The weight of the feature mask loss term in the optimization
                problem.
                Defaults to 1e-4.
            feature_mask_norm_ord (float, optional): The norm p value for the p-norm, which is used on the feature mask.
                Smaller values encourage sparser masks.
                Defaults to 1.
            node_mask_loss_weight (float, optional): The weight of the node mask loss term in the optimization problem.
                Defaults to 0.0.
            node_mask_norm_ord (float, optional): The norm p value for the p-norm, which is used on the feature mask.
                Smaller values encourage sparser masks.
                Defaults to 1.
        """
        super(GNNExplainerOptimizer, self).__init__(**kwargs)
        self.gnn_model = gnn_model
        self._edge_mask_dim = self.gnn_model.get_number_of_edges(
            graph_instance)
        self._feature_mask_dim = self.gnn_model.get_number_of_node_features(
            graph_instance)
        self._node_mask_dim = self.gnn_model.get_number_of_nodes(
            graph_instance)
        self.edge_mask = self.add_weight('edge_mask',
                                         shape=[self._edge_mask_dim, 1],
                                         initializer=ks.initializers.Constant(
                                             value=5.),
                                         dtype=tf.float32,
                                         trainable=True)
        self.feature_mask = self.add_weight('feature_mask',
                                            shape=[self._feature_mask_dim, 1],
                                            initializer=ks.initializers.Constant(
                                                value=5.),
                                            dtype=tf.float32,
                                            trainable=True)
        self.node_mask = self.add_weight('node_mask',
                                         shape=[self._node_mask_dim, 1],
                                         initializer=ks.initializers.Constant(
                                             value=5.),
                                         dtype=tf.float32,
                                         trainable=True)
        self.output_to_explain = gnn_model.predict(graph_instance)

        # Configuration Parameters
        self.edge_mask_loss_weight = edge_mask_loss_weight
        self.edge_mask_norm_ord = edge_mask_norm_ord
        self.feature_mask_loss_weight = feature_mask_loss_weight
        self.feature_mask_norm_ord = feature_mask_norm_ord
        self.node_mask_loss_weight = node_mask_loss_weight
        self.node_mask_norm_ord = node_mask_norm_ord

    def call(self, graph_input, training: bool = False, **kwargs):
        """Call GNN model.

        Args:
            graph_input: Graph input.
            training (bool): If training mode. Default is False.

        Returns:
            tf.tensor: Masked prediction of GNN model.
        """
        edge_mask = self.get_mask("edge")
        feature_mask = self.get_mask("feature")
        node_mask = self.get_mask("node")
        return self.gnn_model.masked_predict(graph_input, edge_mask, feature_mask, node_mask, training=training)

    def train_step(self, data):
        """Train step."""

        graph_input = data[0]

        with tf.GradientTape() as tape:
            y_pred = self(graph_input, training=True)  # Forward pass

            # edge_mask loss
            if self.edge_mask_loss_weight > 0:
                self.add_loss(lambda: tf.norm(tf.sigmoid(
                    self.edge_mask), ord=self.edge_mask_norm_ord) * self.edge_mask_loss_weight)
            # feature_mask loss
            if self.feature_mask_loss_weight > 0:
                self.add_loss(lambda: tf.norm(tf.sigmoid(
                    self.feature_mask), ord=self.feature_mask_norm_ord) * self.feature_mask_loss_weight)
            # node_mask loss
            if self.node_mask_loss_weight > 0:
                self.add_loss(lambda: tf.norm(tf.sigmoid(
                    self.node_mask), ord=self.node_mask_norm_ord) * self.node_mask_loss_weight)

            loss = self.compiled_loss(
                self.output_to_explain, y_pred, regularization_losses=self.losses)

        trainable_vars = []
        if self.edge_mask_loss_weight > 0:
            trainable_vars.append(self.edge_mask)
        if self.feature_mask_loss_weight > 0:
            trainable_vars.append(self.feature_mask)
        if self.node_mask_loss_weight > 0:
            trainable_vars.append(self.node_mask)

        gradients = tape.gradient(loss, trainable_vars)
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))
        self.compiled_metrics.update_state(self.output_to_explain, y_pred)
        return {m.name: m.result() for m in self.metrics}

    def get_mask(self, mask_identifier):
        if mask_identifier == "edge":
            return self._get_mask(self.edge_mask, self.edge_mask_loss_weight)
        elif mask_identifier == "feature":
            return self._get_mask(self.feature_mask, self.feature_mask_loss_weight)
        elif mask_identifier == "node":
            return self._get_mask(self.node_mask, self.node_mask_loss_weight)
        raise Exception("mask_identifier must be 'edge', 'feature' or 'node'")

    def _get_mask(self, mask, weight):
        if weight > 0:
            return tf.sigmoid(mask)
        return tf.ones_like(mask)



import os
import pickle
import numpy as np
import scipy.io
import json
import pandas as pd
from typing import Union
from kgcnn.data.qm import QMDataset
from kgcnn.data.download import DownloadDataset


class QM9MolNetDataset(QMDataset, DownloadDataset):
    r"""Store and process QM9 dataset from `MoleculeNet <https://moleculenet.org/>`__ dataset.

    This is the QM9 dataset as preprocessed from `MoleculeNet <https://moleculenet.org/>`__ with structure and labels.
    See :obj:`kgcnn.data.datasets.QM9Dataset` for documentation and comparison.

    References:

        (1) L. Ruddigkeit, R. van Deursen, L. C. Blum, J.-L. Reymond, Enumeration of 166 billion organic small
            molecules in the chemical universe database GDB-17, J. Chem. Inf. Model. 52, 2864–2875, 2012.
        (2) R. Ramakrishnan, P. O. Dral, M. Rupp, O. A. von Lilienfeld, Quantum chemistry structures and properties
            of 134 kilo molecules, Scientific Data 1, 140022, 2014.

    """

    download_info = {
        "dataset_name": "QM9MolNet",
        "data_directory_name": "qm9_mol_net",
        "download_url": "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/gdb9.tar.gz",
        "download_file_name": 'gdb9.tar.gz',
        "unpack_tar": True,
        "unpack_zip": False,
        "unpack_directory_name": "gdb9"
    }

    def __init__(self, reload: bool = False, verbose: int = 10):
        """Initialize QM8 dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        QMDataset.__init__(self, verbose=verbose, dataset_name="QM9MolNet")
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)

        self.label_names = [
            "A", "B", "C", "mu", "alpha", "homo", "lumo", "gap", "r2", "zpve", "u0", "u298", "h298", "g298",
            "cv", "u0_atom", "u298_atom", "h298_atom", "g298_atom"
        ]
        self.label_units = [
            "GHz", "GHz", "GHz", "D", r"a_0^3", "H", "H", "H", r"a_0^2", "H", "H", "H", "H", "H", r"cal/mol K",
            "kcal/mol", "kcal/mol", "kcal/mol", "kcal/mol"]
        self.dataset_name = "QM9MolNet"
        self.require_prepare_data = False
        self.fits_in_memory = True
        self.verbose = verbose
        self.data_directory = os.path.join(self.data_main_dir, self.data_directory_name, self.unpack_directory_name)
        self.file_name = "gdb9.csv"

        if not os.path.exists(self.file_path):
            original_name = os.path.join(self.data_directory, "gdb9.sdf.csv")
            if os.path.exists(original_name):
                os.rename(original_name, self.file_path)

        if self.require_prepare_data:
            self.prepare_data(overwrite=reload)

        if self.fits_in_memory:
            self.read_in_memory(label_column_name=self.label_names)



from kgcnn.data.datasets.MoleculeNetDataset2018 import MoleculeNetDataset2018


class FreeSolvDataset(MoleculeNetDataset2018):
    r"""Store and process 'FreeSolv' dataset from `MoleculeNet <https://moleculenet.org/>`_ database.
    Class inherits from :obj:`MoleculeNetDataset2018` and downloads dataset on class initialization.
    Compare reference:
    `DeepChem <https://deepchem.readthedocs.io/en/latest/api_reference/moleculenet.html#freesolv-dataset>`__ reading:
    Experimental and calculated hydration free energy of small molecules in water.
    Description in DeepChem reads: 'The FreeSolv dataset is a collection of experimental and calculated hydration
    free energies for small molecules in water, along with their experiemental values. Here, we are using a modified
    version of the dataset with the molecule smile string and the corresponding experimental hydration free energies.'

    Random splitting is recommended for this dataset.

    References:

        (1) Lukasz Maziarka, et al. Molecule Attention Transformer.
            NeurIPS 2019 arXiv:2002.08264v1 [cs.LG].
        (2) Mobley DL, Guthrie JP. FreeSolv: a database of experimental and calculated hydration free energies,
            with input files. J Comput Aided Mol Des. 2014;28(7):711-720. doi:10.1007/s10822-014-9747-x

    """

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize Lipop dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        super(FreeSolvDataset, self).__init__("FreeSolv", reload=reload, verbose=verbose)



import logging
import numpy as np
import tensorflow as tf
import pandas as pd
import os
# import typing as t
from typing import Union, List, Callable, Dict, Optional
# from collections.abc import MutableSequence

from kgcnn.data.utils import save_pickle_file, load_pickle_file, ragged_tensor_from_nested_numpy
from kgcnn.graph.base import GraphDict

# Module logger
logging.basicConfig()
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


class MemoryGraphList(list):
    r"""Class to store a list of graph dictionaries in memory.

    Inherits from a python list. The graph properties are defined by tensor-like (numpy) arrays
    for indices, attributes, labels, symbol etc. in :obj:`GraphDict`, which are the items of the list.
    Access to items via `[]` indexing operator.

    A python list of a single named property can be obtained from each :obj:`GraphDict` in :obj:`MemoryGraphList` via
    :obj:`get` or assigned from a python list via :obj:`set` methods.

    The :obj:`MemoryGraphList` further provides simple map-functionality :obj:`map_list` to apply methods for
    each :obj:`GraphDict`, and to cast properties to tensor with :obj:`tensor`.

    Cleaning the list for missing properties or empty graphs is done with :obj:`clean`.

    .. code-block:: python

        import numpy as np
        from kgcnn.data.base import MemoryGraphList

        data = MemoryGraphList()
        data.empty(1)
        data.set("edge_indices", [np.array([[0, 1], [1, 0]])])
        data.set("node_labels", [np.array([[0], [1]])])
        print(data.get("edge_indices"))
        data.set("node_coordinates", [np.array([[1, 0, 0], [0, 1, 0], [0, 2, 0], [0, 3, 0]])])
        data.map_list("set_range", max_distance=1.5, max_neighbours=10, self_loops=False)
        data.clean("range_indices")  # Returns cleaned graph indices
        print(len(data))
        print(data[0])

    """

    _require_validate = True

    def __init__(self, iterable: list = None):
        r"""Initialize an empty :obj:`MemoryGraphList` instance.

        Args:
            iterable (list, MemoryGraphList): A list or :obj:`MemoryGraphList` of :obj:`GraphDict` items.
        """
        iterable = iterable if iterable is not None else []
        super(MemoryGraphList, self).__init__(iterable)
        self.logger = module_logger
        self.validate()

    def validate(self):
        for i, x in enumerate(self):
            if not isinstance(x, GraphDict):
                self[i] = GraphDict(x)

    def assign_property(self, key: str, value: list):
        """Assign a list of numpy arrays of a property to :obj:`GraphDict` s in this list.

        Args:
            key (str): Name of the property.
            value (list): List of numpy arrays for property `key` .

        Returns:
            self
        """
        if value is None:
            # We could also here remove the key from all graphs.
            return self
        if not isinstance(value, list):
            raise TypeError("Expected type 'list' to assign graph properties.")
        if len(self) == 0:
            self.empty(len(value))
        if len(self) != len(value):
            raise ValueError("Can only store graph attributes from list with same length.")
        for i, x in enumerate(value):
            self[i].assign_property(key, x)
        return self

    def obtain_property(self, key: str) -> Union[List, None]:
        r"""Returns a list with the values of all the graphs defined for the string property name `key`. If none of
        the graphs in the list have this property, returns None.

        Args:
            key (str): The string name of the property to be retrieved for all the graphs contained in this list
        """
        # values for teach of the graphs which make up this list.
        prop_list = [x.obtain_property(key) for x in self]

        # If a certain string property is not set for a GraphDict, it will still return None. Here we check:
        # If all the items for our given property name are None then we know that this property is generally not
        # defined for any of the graphs in the list.
        if all([x is None for x in prop_list]):
            self.logger.warning("Property '%s' is not set on any graph." % key)
            return None

        return prop_list

    def __getitem__(self, item) -> Union[GraphDict, List]:
        # Does not make a copy of the data, as a python list does.
        if isinstance(item, int):
            return super(MemoryGraphList, self).__getitem__(item)
        if isinstance(item, slice):
            return MemoryGraphList(super(MemoryGraphList, self).__getitem__(item))
        if isinstance(item, (list, tuple)):
            return MemoryGraphList([super(MemoryGraphList, self).__getitem__(int(i)) for i in item])
        if isinstance(item, np.ndarray):
            return MemoryGraphList([super(MemoryGraphList, self).__getitem__(int(i)) for i in item])
        raise TypeError("Unsupported type for `MemoryGraphList` items.")

    def __setitem__(self, key, value):
        if not isinstance(value, GraphDict):
            raise TypeError("Require a GraphDict as list item.")
        super(MemoryGraphList, self).__setitem__(key, value)

    def __repr__(self):
        return "<{} [{}]>".format(type(self).__name__, "" if len(self) == 0 else self[0].__repr__() + " ...")

    def append(self, graph):
        assert isinstance(graph, GraphDict), "Must append `GraphDict` to self."
        super(MemoryGraphList, self).append(graph)

    def insert(self, index: int, value) -> None:
        assert isinstance(value, GraphDict), "Must insert `GraphDict` to self."
        super(MemoryGraphList, self).insert(index, value)

    def __add__(self, other):
        assert isinstance(other, MemoryGraphList), "Must add `MemoryGraphList` to self."
        return MemoryGraphList(super(MemoryGraphList, self).__add__(other))

    def copy(self):
        """Copy data in the list."""
        return MemoryGraphList([x.copy() for x in self])

    def empty(self, length: int):
        """Create an empty list in place. Overwrites existing list.

        Args:
            length (int): Length of the empty list.

        Returns:
            self
        """
        if length is None:
            return self
        if length < 0:
            raise ValueError("Length of empty list must be >=0.")
        self.clear()
        for _ in range(length):
            self.append(GraphDict())
        return self

    def update(self, other) -> None:
        assert isinstance(other, MemoryGraphList), "Must update `MemoryGraphList`."
        assert len(other) == len(self), "Length of list for update must match."
        for i in range(len(self)):
            self[i].update(other[i])

    @property
    def length(self):
        """Length of list."""
        return len(self)

    @length.setter
    def length(self, value: int):
        raise ValueError("Can not set length. Please use 'empty()' to initialize an empty list.")

    def _to_tensor(self, item: dict, make_copy=True):
        # TODO: Document this
        if not make_copy:
            self.logger.warning("At the moment always a copy is made for tensor().")
        props: list = self.obtain_property(item["name"])
        is_ragged = item["ragged"] if "ragged" in item else False
        dtype = item["dtype"] if "dtype" in item else None
        # Could add name and shape information if available.
        if is_ragged:
            return ragged_tensor_from_nested_numpy(props, dtype=dtype)
        else:
            return tf.constant(np.array(props), dtype=dtype)

    def tensor(self, items: Union[list, Dict], make_copy=True):
        r"""Make tensor objects from multiple graph properties in list.

        It is recommended to run :obj:`clean` beforehand.

        Args:
            items (list): List of dictionaries that specify graph properties in list via 'name' key.
                The dict-items match the tensor input for :obj:`tf.keras.layers.Input` layers.
                Required dict-keys should be 'name' and 'ragged'.
                Optionally shape information can be included via 'shape' and 'dtype'.
                E.g.: `[{'name': 'edge_indices', 'ragged': True}, {...}, ...]`.
            make_copy (bool): Whether to copy the data. Default is True.

        Returns:
            list: List of Tensors.
        """
        if isinstance(items, dict):
            if all([isinstance(value, dict) for value in items.values()]) and "name" not in items:
                return {key: self._to_tensor(value, make_copy=make_copy) for key, value in items.items()}
            return self._to_tensor(items, make_copy=make_copy)
        elif isinstance(items, (tuple, list)):
            return [self._to_tensor(x, make_copy=make_copy) for x in items]
        else:
            raise TypeError("Wrong type, expected e.g. [{'name': 'edge_indices', 'ragged': True}, {...}, ...]")

    def map_list(self, method: Union[str, Callable], **kwargs):
        r"""Map a method over this list and apply on each :obj:`GraphDict`.
        For :obj:`method` being string, either a class-method or a preprocessor is chosen for backward compatibility.

        .. code-block:: python

            for i, x in enumerate(self):
                method(x, **kwargs)

        Args:
            method (str): Name of the :obj:`GraphDict` method.
            kwargs: Kwargs for `method`.

        Returns:
            self
        """
        # Can add progress info here.
        # Method by name.
        if isinstance(method, str):
            for i, x in enumerate(self):
                # If this is a class method.
                if hasattr(x, method):
                    getattr(x, method)(**kwargs)
                else:
                    # For compatibility names can refer to preprocessors.
                    x.apply_preprocessor(name=method, **kwargs)
        elif isinstance(method, dict):
            raise NotImplementedError("Serialization for method in `map_list` is not yet supported")
        else:
            # For any callable method to map.
            for i, x in enumerate(self):
                method(x, **kwargs)
        return self

    def clean(self, inputs: Union[list, str]):
        r"""Given a list of property names, this method removes all elements from the internal list of
        `GraphDict` items, which do not define at least one of those properties. Meaning, only those graphs remain in
        the list which definitely define all properties specified by :obj:`inputs`.

        Args:
            inputs (list): A list of strings, where each string is supposed to be a property name, which the graphs
                in this list may possess. Within :obj:`kgcnn`, this can be simpy the 'input' category in model
                configuration. In this case, a list of dicts that specify the name of the property with 'name' key.

        Returns:
            invalid_graphs (np.ndarray): A list of graph indices that do not have the required properties and which
                have been removed.
        """
        if isinstance(inputs, str):
            inputs = [inputs]
        invalid_graphs = []
        for item in inputs:
            # If this is a list of dict, which are the config for ks.layers.Input(), we pick 'name'.
            if isinstance(item, dict):
                item_name = item["name"]
            else:
                item_name = item
            props = self.obtain_property(item_name)
            if props is None:
                self.logger.warning("Can not clean property '%s' as it was not assigned to any graph." % item)
                continue
            for i, x in enumerate(props):
                # If property is neither list nor np.array
                if x is None or not hasattr(x, "__getitem__"):
                    self.logger.info("Property '%s' is not defined for graph '%s'." % (item_name, i))
                    invalid_graphs.append(i)
                elif not isinstance(x, np.ndarray):
                    self.logger.info("Property '%s' is not a numpy array for graph '%s'." % (item_name, i))
                    invalid_graphs.append(i)
                elif len(x.shape) > 0:
                    if len(x) <= 0:
                        self.logger.info("Property '%s' is an empty list for graph '%s'." % (item_name, i))
                        invalid_graphs.append(i)
        invalid_graphs = np.unique(np.array(invalid_graphs, dtype="int"))
        invalid_graphs = np.flip(invalid_graphs)  # Need descending order for pop()
        if len(invalid_graphs) > 0:
            self.logger.warning("Found invalid graphs for properties. Removing graphs '%s'." % invalid_graphs)
        else:
            self.logger.info("No invalid graphs for assigned properties found.")
        # Remove from the end via pop().
        for i in invalid_graphs:
            self.pop(int(i))
        return invalid_graphs

    # Alias of internal assign and obtain property.
    set = assign_property
    get = obtain_property


class MemoryGraphDataset(MemoryGraphList):
    r"""Dataset class for lists of graph tensor dictionaries stored on file and fit into memory.

    This class inherits from :obj:`MemoryGraphList` and can be used (after loading and setup) as such.
    It has further information about a location on disk, i.e. a file directory and a file
    name as well as a name of the dataset.

    .. code-block:: python

        import numpy as np
        from kgcnn.data.base import MemoryGraphDataset
        dataset = MemoryGraphDataset(data_directory="", dataset_name="Example")
        # Methods of MemoryGraphList
        dataset.set("edge_indices", [np.array([[1, 0], [0, 1]])])
        dataset.set("edge_labels", [np.array([[0], [1]])])
        dataset.save()
        dataset.load()

    The file directory and file name are used in child classes and in :obj:`save` and :obj:`load` .

    """

    fits_in_memory = True

    def __init__(self,
                 data_directory: str = None,
                 dataset_name: str = None,
                 file_name: str = None,
                 file_directory: str = None,
                 verbose: int = 10,
                 ):
        r"""Initialize a base class of :obj:`MemoryGraphDataset`.

        Args:
            data_directory (str): Full path to directory of the dataset. Default is None.
            file_name (str): Generic filename for dataset to read into memory like a 'csv' file. Default is None.
            file_directory (str): Name or relative path from :obj:`data_directory` to a directory containing sorted
                files. Default is None.
            dataset_name (str): Name of the dataset. Important for naming and saving files. Default is None.
            verbose (int): Logging level. Default is 10.
        """
        super(MemoryGraphDataset, self).__init__()
        # For logging.
        self.logger = logging.getLogger("kgcnn.data." + dataset_name) if dataset_name is not None else module_logger
        self.logger.setLevel(verbose)
        # Information on location of dataset on file. Data directory must be filepath.
        self.data_directory = data_directory
        self.file_name = file_name
        self.file_directory = file_directory
        self.dataset_name = dataset_name
        # Data Frame for labels and graph names.
        self.data_frame = None
        self.data_keys = None
        self.data_unit = None

    def _verify_data_directory(self) -> Union[str, None]:
        r"""Utility function that checks if `data_directory` is set correctly."""
        if self.data_directory is None:
            self.warning("Data directory is not set.")
            return None
        if not os.path.exists(os.path.realpath(self.data_directory)):
            self.error("Data directory does not exist.")
        return self.data_directory

    @property
    def file_path(self):
        r"""Construct filepath from 'file_name' given in `init`."""
        self._verify_data_directory()
        if self.file_name is None:
            self.warning("Can not determine file path, missing `file_name`.")
            return None
        return os.path.join(self.data_directory, self.file_name)

    @property
    def file_directory_path(self):
        r"""Construct file-directory path from 'data_directory' and 'file_directory' given in `init`."""
        self._verify_data_directory()
        if self.file_directory is None:
            self.warning("Can not determine file directory, missing `file_directory`.")
            return None
        return os.path.join(self.data_directory, self.file_directory)

    def info(self, *args, **kwargs):
        """Pass information to class' logger instance."""
        self.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        """Pass warning to class' logger instance."""
        self.logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        """Pass error to class' logger instance."""
        self.logger.error(*args, **kwargs)

    def save(self, filepath: str = None):
        r"""Save all graph properties to python dictionary as pickled file. By default, saves a file named
        :obj:`dataset_name.kgcnn.pickle` in :obj:`data_directory` .

        Args:
            filepath (str): Full path of output file. Default is None.
        """
        if filepath is None:
            filepath = os.path.join(self.data_directory, self.dataset_name + ".kgcnn.pickle")
        self.info("Pickle dataset...")
        save_pickle_file([x.to_dict() for x in self], filepath)
        return self

    def load(self, filepath: str = None):
        r"""Load graph properties from a pickled file. By default, loads a file named
        :obj:`dataset_name.kgcnn.pickle` in :obj:`data_directory` .

        Args:
            filepath (str): Full path of input file.
        """
        if filepath is None:
            filepath = os.path.join(self.data_directory, self.dataset_name + ".kgcnn.pickle")
        self.info("Load pickled dataset...")
        in_list = load_pickle_file(filepath)
        self.clear()
        for x in in_list:
            self.append(GraphDict(x))
        return self

    def read_in_table_file(self, file_path: str = None, **kwargs):
        r"""Read a data frame in :obj:`data_frame` from file path. By default, uses :obj:`file_name` and pandas.
        Checks for a '.csv' file and then for Excel file endings. Meaning the file extension of file_path is ignored
        but must be any of the following '.csv', '.xls', '.xlsx', '.odt'.

        Args:
            file_path (str): File path to table file. Default is None.
            kwargs: Kwargs for pandas :obj:`read_csv` function.

        Returns:
            self
        """
        if file_path is None:
            file_path = os.path.join(self.data_directory, self.file_name)

        # TODO: Better determine file-type from file ending and just test all supported types otherwise.
        # file_extension_given = os.path.splitext(file_path)[1]
        file_path_base = os.path.splitext(file_path)[0]

        for file_extension in [".csv"]:
            if os.path.exists(file_path_base + file_extension):
                self.data_frame = pd.read_csv(file_path_base + file_extension, **kwargs)
                return self
        for file_extension in [".xls", ".xlsx", ".xlsm", ".xlsb", ".odf", ".ods", ".odt"]:
            if os.path.exists(file_path_base + file_extension):
                self.data_frame = pd.read_excel(file_path_base + file_extension, **kwargs)
                return self

        self.warning("Unsupported data extension of '%s' for table file." % file_path)
        return self

    def assert_valid_model_input(self, hyper_input: Union[list, dict], raise_error_on_fail: bool = True):
        r"""Check whether dataset has graph properties (in tensor format) requested by model input.

        The list :obj:`hyper_input` that defines model input match interface to hyperparameter.
        The model input is set up by a list of layer configs for the keras :obj:`Input` layer.
        The list must contain dictionaries for each model input with "name" and "shape" keys.

        .. code-block:: python

            hyper_input = [
                {"shape": [None, 8710], "name": "node_attributes", "dtype": "float32", "ragged": True},
                {"shape": [None, 1], "name": "edge_weights", "dtype": "float32", "ragged": True},
                {"shape": [None, 2], "name": "edge_indices", "dtype": "int64", "ragged": True}
            ]

        Args:
            hyper_input (list): List of properties that need to be available to a model for training.
            raise_error_on_fail (bool): Whether to raise an error if assertion failed.
        """
        dataset = self

        def message_error(msg):
            if raise_error_on_fail:
                raise ValueError(msg)
            else:
                dataset.error(msg)

        def message_warning(msg):
            dataset.warning(msg)

        if isinstance(hyper_input, dict):
            if "name" in hyper_input and "shape" in hyper_input:
                # Single model input that has not been properly passed as list.
                # Assume here one does not name model output name and shape.
                hyper_input = [hyper_input]
            else:
                # In principle keras also accepts a dictionary for model inputs. Just cast to list here.
                hyper_input = list(hyper_input.values())

        # Check if we have List[dict].
        for x in hyper_input:
            if not isinstance(x, dict):
                message_error(
                    "Wrong type of list item in `assert_valid_model_input`. Found '%s' but must be `dict` ." % type(x))

        for x in hyper_input:
            if "name" not in x:
                message_error("Can not infer name from '%s' for model input." % x)
            data = [dataset[i].obtain_property(x["name"]) for i in range(len(dataset))]
            prop_in_data = [y is None for y in data]
            if all(prop_in_data):
                message_error("Property %s is not defined for any graph in list. Please check property." % x["name"])
            if any(prop_in_data):
                message_warning("Property %s is not defined for all graphs in list. Please run clean()." % x["name"])

            # we also will check shape here but only with first element.
            if hasattr(data[0], "shape") and "shape" in x:
                shape_element = data[0].shape
                shape_input = x["shape"]
                if len(shape_input) != len(shape_element):
                    message_error(
                        "Mismatch in rank for model input {} vs. {}".format(shape_element, shape_input))
                for i, dim in enumerate(shape_input):
                    if dim is not None:
                        if shape_element[i] != dim:
                            message_error(
                                "Mismatch in shape for model input {} vs. {}".format(shape_element, shape_input))
            else:
                message_error("Can not check shape for '%s'." % x["name"])
        return

    def collect_files_in_file_directory(self, file_column_name: str = None, table_file_path: str = None,
                                        read_method_file: Callable = None, update_counter: int = 1000,
                                        append_file_content: bool = True,
                                        read_method_return_list: bool = False) -> list:
        r"""Utility function to collect single files in :obj:`file_directory` by names in CSV table file.

        Args:
            file_column_name (str): Name of the column in Table file that holds list of file names.
            table_file_path (str): Path to table file. Can be None. Default is None.
            read_method_file (Callable): Callable read-file method to return (processed) file content.
            update_counter (int): Loop counter to show progress. Default is 1000.
            append_file_content (bool): Whether to append or add return of :obj:`read_method_file`.
            read_method_return_list (bool): Whether :obj:`read_method_file` returns list of items or the item itself.

        Returns:
            list: File content loaded from single files.
        """
        self.read_in_table_file(table_file_path)
        if self.data_frame is None:
            raise FileNotFoundError("Can not find '.csv' table path '%s'." % table_file_path)

        if file_column_name is None:
            raise ValueError("Please specify column for '.csv' file which contains file names.")

        if file_column_name not in self.data_frame.columns:
            raise ValueError(
                "Can not find file names of column '%s' in '%s'" % (file_column_name, self.data_frame.columns))

        name_file_list = self.data_frame[file_column_name].values
        num_files = len(name_file_list)

        if not os.path.exists(self.file_directory_path):
            raise ValueError("No file directory found at '%s'." % self.file_directory_path)

        self.info("Read %s single files." % num_files)
        out_list = []
        for i, x in enumerate(name_file_list):
            # Only one file per path
            file_loaded = read_method_file(os.path.join(self.file_directory_path, x))
            if append_file_content:
                if read_method_return_list:
                    out_list.append(file_loaded[0])
                else:
                    out_list.append(file_loaded)
            else:
                if read_method_return_list:
                    out_list += file_loaded
                else:
                    out_list += [file_loaded]
            if i % update_counter == 0:
                self.info("... Read {0} file {1} from {2}".format(os.path.splitext(x)[1], i, num_files))

        return out_list

    def set_methods(self, method_list: List[dict]) -> None:
        r"""Apply a list of serialized class-methods on the dataset.

        This can extend the config-serialization scheme in :obj:`kgcnn.utils.serial`.

        .. code-block:: python

            for method_item in method_list:
                for method, kwargs in method_item.items():
                    if hasattr(self, method):
                        getattr(self, method)(**kwargs)

        Args:
            method_list (list): A list of dictionaries that specify class methods. The `dict` key denotes the method
                and the value must contain `kwargs` for the method

        Returns:
            None.
        """
        for method_item in method_list:
            for method, kwargs in method_item.items():
                if hasattr(self, method):
                    getattr(self, method)(**kwargs)
                else:
                    self.error("Class does not have method '%s'." % method)

    def get_split_indices(self,
                          name: str = "kfold",
                          return_as_train_test: bool = True,
                          shuffle: bool = True,
                          seed: int = None
                          ):
        """Gather split ids from split graph property and return k-fold splits.

        Args:
            name (str): Name of property containing split assignment. Default is "kfold".
            return_as_train_test (bool): Whether to return the splits as train test indices. Default is True.
            shuffle (bool): Whether to shuffle splits. Default is True.
            seed (int): Random seed for shuffle. Default is None.

        Returns:
            list: List of splits.
        """
        split_indices = [[]]

        def check_and_extend_splits(to_split):
            if to_split - len(split_indices) + 1 > 0:
                for _ in range(to_split - len(split_indices) + 1):
                    split_indices.append([])
        graphs = self.obtain_property(name)
        for i, s in enumerate(graphs):
            if s is None:
                self.error("Split assignment on graph '%s' is not defined." % i)
                continue
            for x in s:
                check_and_extend_splits(x)
                split_indices[int(x)].append(i)

        split_indices = [np.array(x, dtype="int") for x in split_indices]

        if shuffle:
            np.random.seed(seed)
            for x in split_indices:
                np.random.shuffle(x)

        if not return_as_train_test:
            return split_indices

        train_test = []

        for i in range(len(split_indices)):
            train = np.concatenate([x for j, x in enumerate(split_indices) if j != i], axis=0)
            test = split_indices[i]
            train_test.append([train, test])

        return train_test

    def get_train_test_indices(self,
                               train: str = "train",
                               test: str = "test",
                               valid: Optional[str] = None,
                               split_index: Union[int, list] = 1,
                               shuffle: bool = False,
                               seed: int = None
                               ) -> List[List[np.ndarray]]:
        """
        Get train and test indices from graph list.
        The 'train' and 'test' properties must be set on the graph, and optionally an additional property
        for the validation split may be present. All of these properties may have either of the following
        values:

        - The property is a boolean integer value indicating whether the corresponding element of the
          dataset belongs to that part of the split (train / test)
        - The property is a list containing integer split indices, where each split index present within
          that list implies that the corresponding dataset element is part of that particular split.
          In this case the ``split_index`` parameter may also be a list of split indices that specifies
          for which of these split indices the train test index split is to be returned by this method.

        The return value of this method is a list with the same length as the ``split_index`` parameter,
        which by default will be 1.

        Args:
            train (str): Name of graph property that has train split assignment. Defaults to 'train'.
            test (str): Name of graph property that has test split assignment. Defaults to 'test'.
            valid (str): Name of graph property that has validation assignment. Defaults to None.
            split_index (int, list): Split index to get indices for. Can also be list.
            shuffle (bool): Whether to shuffle splits. Default is True.
            seed (int): Random seed for shuffle. Default is None.

        Returns:
            list: List of tuples (or triples) of train, test, (validation) split indices.
        """
        out_indices = []
        if not isinstance(split_index, (list, tuple)):
            split_index_list: List[int] = [split_index]
        else:
            split_index_list: List[int] = split_index

        for split_index in split_index_list:

            # This list will itself contain numpy arrays which are filled with graph indices of the dataset
            # each element of this list will correspond to one property name (train, test...)
            graph_index_split_list: List[np.ndarray] = []

            for property_name in [train, test, valid]:

                # It may be that we only seek train and test indices and not validation indices, in that
                # case the property name for validation is None, in which case we want to skip
                if property_name is None:
                    continue

                # This list will contain all the indices of the dataset elements (graphs) which are
                # associated with the current iteration's split index for the current iteration's
                # property name (train, test...)
                graph_index_list: List[int] = []

                # "obtain_property" returns a list which contains only the property values corresponding to
                # the given property name for each graph inside the dataset in the same order.
                # In this case, this is supposed to be a split list, which is a list that contains integer
                # indices, each representing one particular dataset split. The split list of each graph
                # only contains those split indices to which that graph is associated.
                split_prop: List[List[int]] = self.obtain_property(property_name)
                for index, split_list in enumerate(split_prop):
                    if split_list is not None:
                        if split_index in split_list:
                            graph_index_list.append(index)

                graph_index_array = np.array(graph_index_list)
                graph_index_split_list.append(graph_index_array)

            if shuffle:
                np.random.seed(seed)
                for graph_index_array in graph_index_split_list:
                    np.random.shuffle(graph_index_array)

            out_indices.append(graph_index_split_list)

        return out_indices

    def relocate(self, data_directory: str = None, file_name: str = None, file_directory: str = None):
        """Change file information. Does not copy files on disk!

        Args:
            data_directory (str): Full path to directory of the dataset. Default is None.
            file_name (str): Generic filename for dataset to read into memory like a 'csv' file. Default is None.
            file_directory (str): Name or relative path from :obj:`data_directory` to a directory containing sorted
                files. Default is None.

        Returns:
            self
        """
        self.data_directory = data_directory
        self.file_name = file_name
        self.file_directory = file_directory
        return self


MemoryGeometricGraphDataset = MemoryGraphDataset



import os
import numpy as np
from ase.db import connect

from kgcnn.data.base import MemoryGraphDataset
from kgcnn.data.download import DownloadDataset


class ISO17Dataset(DownloadDataset, MemoryGraphDataset):
    r"""Dataset 'ISO17' with molecules randomly taken from QM9 dataset [1] with a fixed composition of atoms (C7O2H10).
    They were arranged in different chemically valid structures and is an extension of the isomer MD data used in [2].

    Information below and the dataset itself is copied and downloaded from `<http://quantum-machine.org/datasets/>`_ .
    The database consist of 129 molecules each containing 5000 conformational geometries,
    energies and forces with a resolution of 1 femto-second in the molecular dynamics trajectories.
    The database was generated from molecular dynamics simulations using the Fritz-Haber Institute ab initio simulation
    package (FHI-aims)[3]. The simulations in ISO17 were carried out using the standard quantum chemistry computational
    method  density functional theory (DFT) in the generalized gradient approximation (GGA) with the
    Perdew-Burke-Ernzerhof (PBE) functional[4] and the Tkatchenko-Scheffler (TS) van der Waals correction method [5].
    The dataset is stored in ASE sqlite format with the total energy in eV and forces in eV/Ang.
    Download-url: `<http://quantum-machine.org/datasets/iso17.tar.gz>`_ .

    .. code-block:: python

        from ase.db import connect
        with connect(path_to_db) as conn:
           for row in conn.select(limit=10):
               print(row.toatoms())
               print(row['total_energy'])
               print(row.data['atomic_forces'])

    References:

        (1) R. Ramakrishnan et al. Quantum chemistry structures and properties of 134 kilo molecules.
            Scientific Data, 1. (2014) `<https://www.nature.com/articles/sdata201422>`_ .
        (2) Schütt, K. T. et al. Quantum-chemical insights from deep tensor neural networks.
            Nature Communications, 8, 13890. (2017) `<https://www.nature.com/articles/ncomms13890>`_ .
        (3) Blum, V. et al. Ab Initio Molecular Simulations with Numeric Atom-Centered Orbitals.
            Comput. Phys. Commun. 2009, 180 (11), 2175–2196
            `<https://www.sciencedirect.com/science/article/pii/S0010465509002033>`_ .
        (4) Perdew, J. P. et al. Generalized Gradient Approximation Made Simple. Phys. Rev. Lett. 1996, 77 (18),
            3865–3868 `<https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.77.3865>`_.
        (5) Tkatchenko, A. et al. Accurate Molecular Van Der Waals Interactions from Ground-State Electron Density and
            Free-Atom Reference Data. Phys. Rev. Lett. 2009, 102 (7), 73005
            `<https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.102.073005>`_.
        (6) Schütt, K. T., et al. SchNet: A continuous-filter convolutional neural network for modeling quantum
            interactions. Advances in Neural Information Processing System (accepted). 2017.
            `<https://arxiv.org/abs/1706.08566>`_
    """

    download_info = {
        "dataset_name": "ISO17",
        "data_directory_name": "ISO17",
        "download_url": "http://quantum-machine.org/datasets/iso17.tar.gz",
        "download_file_name": 'iso17.tar.gz',
        "unpack_tar": True,
        "unpack_zip": False,
        "unpack_directory_name": "iso17"
    }

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize full :obj:`ISO17Dataset` dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        # Use default base class init()
        self.data_keys = None

        MemoryGraphDataset.__init__(self, dataset_name="ISO17", verbose=verbose)
        DownloadDataset.__init__(self, **self.download_info, reload=reload, verbose=verbose)

        self.file_name = ["reference.db", "reference_eq.db", "test_within.db", "test_other.db", "test_eq.db"]
        self.data_directory = os.path.join(
            self.data_main_dir, self.data_directory_name, self.unpack_directory_name, "iso17")
        self.file_directory = self.data_directory

        if self.fits_in_memory:
            self.read_in_memory()

    def read_in_memory(self):
        r"""Load complete :obj:`ISO17Dataset` data into memory. Additionally, the different train and validation
        properties are set according to `http://quantum-machine.org/datasets/ <http://quantum-machine.org/datasets/>`_.

        The data is partitioned as used in the SchNet paper [6]:

            - reference.db: 80% of steps of 80% of MD trajectories
            - reference_eq.db: equilibrium conformations of those molecules
            - test_within.db: remaining 20% unseen steps of reference trajectories
            - test_other.db: remaining 20% unseen MD trajectories
            - test_eq.db: equilibrium conformations of test trajectories

        Where 'reference.db' and 'reference_eq.db' have 'train' property with index 0, 1, respectively, and
        'test_within', 'test_other', 'test_eq' have 'test' property 0, 1, 2, respectively.
        The original validation geometries are noted by 'valid' property with index 0.
        Use :obj:`get_train_test_indices` for reading out split indices.
        """

        data = {"formula": [], "numbers": [], "symbols": [], "positions": [],
                "atomic_forces": [], "total_energy": [], "id": [], "train": [], "test": []}
        # Add them in the order as given in doc string. 'reference' prefix are training data.
        for db_name, train, test in [("reference.db", 0, None), ("reference_eq.db", 1, None),
                                     ("test_within.db", None, 0), ("test_other.db", None, 1), ("test_eq.db", None, 2)]:
            with connect(os.path.join(self.data_directory, db_name)) as conn:
                for row in conn.select():
                    # print(str(row.toatoms()))  # The ase atoms object.
                    data["numbers"].append(row.numbers)
                    data["id"].append(row.id)
                    data["formula"].append(str(row.toatoms().symbols))
                    data["positions"].append(row.positions)
                    data["symbols"].append(row.symbols)
                    data["total_energy"].append(row['total_energy'])
                    data["atomic_forces"].append(row.data['atomic_forces'])
                    data["train"].append(train)
                    data["test"].append(test)

        for name, values in data.items():
            self.assign_property(name, values)

        # Validation indices are for the first part that is 'reference.db' which are at the beginning of the dataset.
        # Can be simply added by index.
        with open(os.path.join(self.data_directory, "validation_ids.txt")) as f:
            valid_indices = [int(x.strip()) for x in f.readlines()]
        for i in valid_indices:
            self[i-1].update({"valid": np.array(0)})
        return self



import unittest

import numpy as np

from kgcnn.xai.utils import flatten_importances_list


class TestFunctions(unittest.TestCase):

    def test_flatten_importances_list_basically_works(self):
        importances_list = [
            np.array([
                [0, 1],
                [1, 0]
            ]),
            np.array([
                [0, 0, 0],
                [1, 0, 0]
            ])
        ]
        expected = [0, 1, 1, 0, 0, 0, 0, 1, 0, 0]

        result = flatten_importances_list(importances_list)
        self.assertListEqual(expected, result)



from kgcnn.data.datasets.MatBenchDataset2020 import MatBenchDataset2020


class MatProjectLogKVRHDataset(MatBenchDataset2020):
    """Store and process :obj:`MatProjectLogKVRHDataset` from `MatBench <https://matbench.materialsproject.org/>`__
    database. Name within Matbench: 'matbench_log_kvrh'.

    Matbench v0.1 test dataset for predicting DFT log10 VRH-average bulk modulus from structure.
    Adapted from Materials Project database. Removed entries having a formation energy (or energy above the convex hull)
    more than 150meV and those having negative G_Voigt, G_Reuss, G_VRH, K_Voigt, K_Reuss, or K_VRH and those failing
    G_Reuss <= G_VRH <= G_Voigt or K_Reuss <= K_VRH <= K_Voigt and those containing noble gases.
    Retrieved April 2, 2019. For benchmarking w/ nested cross validation, the order of the dataset must be identical
    to the retrieved data; refer to the Automatminer/Matbench publication for more details.

        * Number of samples: 10987
        * Task type: regression
        * Input type: structure

    """

    def __init__(self, reload=False, verbose: int = 10):
        r"""Initialize 'matbench_mp_e_form' dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
        """
        # Use default base class init()
        super(MatProjectLogKVRHDataset, self).__init__("matbench_log_kvrh", reload=reload, verbose=verbose)
        self.label_names = "log10(K_VRH) "
        self.label_units = "GPa"



"""
Module containing utility methods for the command line interface (CLI).
"""
import click

CHECK_MARK = '✓'


# == CLICK SPECIFIC UTILS ==

def echo_info(content: str, verbose: bool = True):
    if verbose:
        click.secho(f'... {content}')


def echo_success(content: str, verbose: bool = True):
    if verbose:
        click.secho(f'[{CHECK_MARK}] {content}', fg='green')


def echo_error(content: str, verbose: bool = True):
    if verbose:
        click.secho(f'[!] {content}', fg='red')



import numpy as np
import os
import logging
from kgcnn.mol.base import MolGraphInterface
from openbabel import openbabel

# Module logger.
logging.basicConfig()
module_logger = logging.getLogger(__name__)
module_logger.setLevel(logging.INFO)


# OpenBabel can not be installed with pip but conda only. It is an optional dependency at the moment.
# There was a problem with OpenBabel not finding the 'BABEL_DATADIR'. One can set it manually with
# python `os.environ['BABEL_DATADIR'] = ...` .
if "BABEL_DATADIR" not in os.environ:
    module_logger.warning(
        "System variable 'BABEL_DATADIR' is not set. Please set `os.environ['BABEL_DATADIR'] = ...` manually.")


class MolecularGraphOpenBabel(MolGraphInterface):
    r"""A graph object representing a strict molecular graph, e.g. only chemical bonds.
    This class is an interface to :obj:`OBMol` class to retrieve graph properties.

    .. code-block:: python

        import numpy as np
        from kgcnn.mol.graph_babel import MolecularGraphOpenBabel
        mg = MolecularGraphOpenBabel()
        mg.from_smiles("CC(C)C(C(=O)O)N")
        mg.add_hs()
        mg.make_conformer()
        mg.optimize_conformer()
        mg.compute_partial_charges()
        print(MolecularGraphOpenBabel.atom_fun_dict.keys(), MolecularGraphOpenBabel.bond_fun_dict.keys())
        print(mg.node_coordinates)
        print(mg.edge_indices)
        print(mg.node_attributes(properties=["NumBonds", "GasteigerCharge"], encoder={}))

    """

    atom_fun_dict = {
        "IsInRing": lambda atom: atom.IsInRing(),
        "IsPeriodic": lambda atom: atom.IsPeriodic(),
        "IsHetAtom": lambda atom: atom.IsHetAtom(),
        "IsAxial": lambda atom: atom.IsAxial(),
        "IsMetal": lambda atom: atom.IsMetal(),
        "IsAmideNitrogen": lambda atom: atom.IsAmideNitrogen(),
        "IsAromatic": lambda atom: atom.IsAromatic(),
        "IsAromaticNOxide": lambda atom: atom.IsAromaticNOxide(),
        "IsCarboxylOxygen": lambda atom: atom.IsCarboxylOxygen(),
        "IsChiral": lambda atom: atom.IsChiral(),
        "IsHbondAcceptor": lambda atom: atom.IsHbondAcceptor(),
        "IsHbondAcceptorSimple": lambda atom: atom.IsHbondAcceptorSimple(),
        "IsHbondDonor": lambda atom: atom.IsHbondDonor(),
        "IsHbondDonorH": lambda atom: atom.IsHbondDonorH(),
        "IsHeteroatom": lambda atom: atom.IsHeteroatom(),
        "IsInRingSize6": lambda atom: atom.IsInRingSize(6),
        "IsInRingSize5": lambda atom: atom.IsInRingSize(5),
        "IsNitroOxygen": lambda atom: atom.IsNitroOxygen(),
        "IsNonPolarHydrogen": lambda atom: atom.IsNonPolarHydrogen(),
        "IsPhosphateOxygen": lambda atom: atom.IsPhosphateOxygen(),
        "IsPolarHydrogen": lambda atom: atom.IsPolarHydrogen(),
        "IsSulfateOxygen": lambda atom: atom.IsSulfateOxygen(),
        "Visit": lambda atom: atom.Visit,
        "Isotope": lambda atom: atom.GetIsotope(),
        "Data": lambda bond: bond.GetData(),
        "X": lambda bond: bond.GetX(),
        "Y": lambda bond: bond.GetY(),
        "Z": lambda bond: bond.GetZ(),
        "ExplicitValence": lambda bond: bond.GetExplicitValence(),
        "TotalDegree": lambda bond: bond.GetTotalDegree(),
        "AtomicMass": lambda bond: bond.GetAtomicMass(),
        "AtomicNum": lambda bond: bond.GetAtomicNum(),
        "Coordinate": lambda bond: bond.GetCoordinate(),
        "CoordinateIdx": lambda bond: bond.GetCoordinateIdx(),
        "ExactMass": lambda bond: bond.GetExactMass(),
        "ExplicitDegree": lambda bond: bond.GetExplicitDegree(),
        "FormalCharge": lambda bond: bond.GetFormalCharge(),
        "HeteroDegree": lambda bond: bond.GetHeteroDegree(),
        "HvyDegree": lambda bond: bond.GetHvyDegree(),
        "Hyb": lambda bond: bond.GetHyb(),
        "ImplicitHCount": lambda bond: bond.GetImplicitHCount(),
        "Index": lambda bond: bond.GetIndex(),
        "PartialCharge": lambda bond: bond.GetPartialCharge(),
        "Residue": lambda bond: bond.GetResidue(),
        "SpinMultiplicity": lambda bond: bond.GetSpinMultiplicity(),
        "Title": lambda bond: bond.GetTitle(),
        "TotalValence": lambda bond: bond.GetTotalValence(),
        "Type": lambda bond: bond.GetType(),
        "Vector": lambda bond: bond.GetVector(),
        "HasResidue": lambda bond: bond.HasResidue(),
        "HasDoubleBond": lambda bond: bond.HasDoubleBond(),
        "HasSingleBond": lambda bond: bond.HasSingleBond(),
        "HasAromaticBond": lambda bond: bond.HasAromaticBond(),
        "HasAlphaBetaUnsat": lambda bond: bond.HasAlphaBetaUnsat(),
        "HasBondOfOrder3": lambda bond: bond.HasBondOfOrder(3),
        "HasBondOfOrder2": lambda bond: bond.HasBondOfOrder(2),
        "HasBondOfOrder1": lambda bond: bond.HasBondOfOrder(1),
        "HasNonSingleBond": lambda bond: bond.HasNonSingleBond()
    }
    bond_fun_dict = {
        "BondOrder": lambda bond: bond.GetBondOrder(),
        "IsAromatic": lambda bond: bond.IsAromatic(),
        "IsInRing": lambda bond: bond.IsInRing(),
        "Idx": lambda bond: bond.GetIdx(),
        "Id": lambda bond: bond.GetId(),
        "BeginAtom": lambda bond: bond.GetBeginAtom(),
        "BeginAtomIdx": lambda bond: bond.GetBeginAtomIdx(),
        "EndAtom": lambda bond: bond.GetEndAtom(),
        "EndAtomIdx": lambda bond: bond.GetEndAtomIdx(),
        "Flags": lambda bond: bond.GetFlags(),
        "Parent": lambda bond: bond.GetParent(),
        "Length": lambda bond: bond.GetLength(),
        "EquibLength": lambda bond: bond.GetEquibLength(),
        "Aromatic": lambda bond: bond.Aromatic,
        "CisOrTrans": lambda bond: bond.CisOrTrans,
        "IsHash": lambda bond: bond.IsHash(),
        "IsAmide": lambda bond: bond.IsAmide(),
        "IsEster": lambda bond: bond.IsEster(),
        "IsCarbonyl": lambda bond: bond.IsCarbonyl(),
        "IsCisOrTrans": lambda bond: bond.IsCisOrTrans(),
        "IsClosure": lambda bond: bond.IsClosure(),
        "IsDoubleBondGeometry": lambda bond: bond.IsDoubleBondGeometry(),
        "IsPeriodic": lambda bond: bond.IsPeriodic(),
        "IsPrimaryAmide": lambda bond: bond.IsPrimaryAmide(),
        "IsTertiaryAmide": lambda bond: bond.IsTertiaryAmide(),
        "IsWedge": lambda bond: bond.IsWedge(),
        "IsWedgeOrHash": lambda bond: bond.IsWedgeOrHash(),
        "Visit": lambda bond: bond.Visit,
    }
    mol_fun_dict = {
        "NumAtoms": lambda m: m.NumAtoms(),
        "NumBonds": lambda m: m.NumBonds(),
        "ExactMass": lambda m: m.GetExactMass(),
        "TotalCharge": lambda m: m.GetTotalCharge(),
    }

    def __init__(self, mol=None, make_directed: bool = False):
        """Set the mol attribute for composition. This mol instances will be the backends molecule class.

        Args:
            mol (openbabel.OBMol): OpenBabel molecule.
            make_directed (bool): Whether the edges are directed. Default is False.
        """
        super().__init__(mol=mol, make_directed=make_directed)
        self.mol = mol

    def make_conformer(self, **kwargs):
        r"""Make conformer for mol-object.

        Args:
            kwargs: Not used.

        Returns:
            bool: Whether conformer generation was successful
        """
        if self.mol is None:
            return False
        builder = openbabel.OBBuilder()
        build_okay = builder.Build(self.mol)
        return build_okay

    def optimize_conformer(self, force_field="mmff94", steps=100, **kwargs):
        r"""Optimize conformer. Requires an initial conformer. See :obj:`make_conformer`.

        Args:
            force_field (str): Force field type.
            steps (int): Number of iteration steps.
            kwargs: Kwargs for SteepestDescent.

        Returns:
            bool: Whether conformer optimization was successful.
        """
        if self.mol is None:
            return False
        ff = openbabel.OBForceField.FindType(force_field)
        ff_setup_okay = ff.Setup(self.mol)
        ff.SteepestDescent(steps, **kwargs)  # defaults are 50-500 in pybel
        ff.GetCoordinates(self.mol)
        return ff_setup_okay

    def add_hs(self, **kwargs):
        """Add Hydrogen."""
        self.mol.AddHydrogens(**kwargs)

    def remove_hs(self, **kwargs):
        """Remove Hydrogen."""
        self.mol.DeleteHydrogens(**kwargs)

    def compute_partial_charges(self, method="gasteiger", **kwargs):
        """Compute partial charges.

        Args:
            method (str): Name of charge model.
            kwargs: Not used.

        Returns:
            bool: Compute charges return value.
        """
        mol = self.mol
        ob_charge_model = openbabel.OBChargeModel.FindType(method)
        return ob_charge_model.ComputeCharges(mol)

    def from_smiles(self, smile: str, sanitize: bool = True):
        """Make molecule from smile.

        Args:
            smile (str): Smile string for the molecule.
            sanitize (bool): Whether to sanitize molecule.
        """
        ob_conversion = openbabel.OBConversion()
        ob_conversion.SetInFormat("smiles")
        self.mol = openbabel.OBMol()
        ob_conversion.ReadString(self.mol, smile)
        return self

    def to_smiles(self):
        """Return a smile string representation of the mol instance.

        Returns:
            smile (str): Smile string.
        """
        ob_conversion = openbabel.OBConversion()
        ob_conversion.SetOutFormat("smiles")
        return ob_conversion.WriteString(self.mol)

    def from_mol_block(self, mol_block: str, keep_hs: bool = True, sanitize: bool = True):
        """Set mol-instance from a string representation containing coordinates and bond information that is MDL mol
        format equivalent.

        Args:
            mol_block (str): Mol-block representation of a molecule.
            sanitize (bool): Whether to sanitize the mol-object.
            keep_hs (bool): Whether to keep hydrogen.

        Returns:
            self
        """
        ob_conversion = openbabel.OBConversion()
        ob_conversion.SetInFormat("mol")
        self.mol = openbabel.OBMol()
        ob_conversion.ReadString(self.mol, mol_block)
        if self.mol.HasHydrogensAdded() and not keep_hs:
            self.mol.DeleteHydrogens()
        return self

    def from_xyz(self, xyz_string):
        """Setting mol-instance from an external xyz-string. Does not add hydrogen or makes conformers.

        Args:
            xyz_string:

        Returns:
            self
        """
        ob_conversion = openbabel.OBConversion()
        ob_conversion.SetInFormat("xyz")
        self.mol = openbabel.OBMol()
        ob_conversion.ReadString(self.mol, xyz_string)
        return self

    def to_mol_block(self):
        """Make a more extensive string representation containing coordinates and bond information from self.

        Returns:
            mol_block (str): Mol-block representation of a molecule.
        """
        ob_conversion = openbabel.OBConversion()
        ob_conversion.SetOutFormat("mol")
        return ob_conversion.WriteString(self.mol)

    @property
    def node_number(self):
        """Return list of node numbers which is the atomic number of atoms in the molecule"""
        atom_num = []
        for i in range(self.mol.NumAtoms()):
            ats = self.mol.GetAtomById(i)
            # ats = mol.GetAtom(i+1)
            atom_num.append(ats.GetAtomicNum())
        return atom_num

    @property
    def node_symbol(self):
        """Return a list of atomic symbols of the molecule."""
        atom_type = []
        for i in range(self.mol.NumAtoms()):
            ats = self.mol.GetAtomById(i)
            # ats = mol.GetAtom(i+1)
            atom_type.append(ats.GetType())
        return atom_type

    @property
    def node_coordinates(self):
        """Return a list of atomic coordinates of the molecule."""
        xyz = []
        for i in range(self.mol.NumAtoms()):
            ats = self.mol.GetAtomById(i)
            # ats = mol.GetAtom(i+1)
            xyz.append([ats.GetX(), ats.GetY(), ats.GetZ()])
        if len(xyz) <= 0:
            return
        return np.array(xyz)

    @property
    def edge_indices(self):
        """Return a list of edge indices of the molecule."""
        bond_idx = []
        for i in range(self.mol.NumBonds()):
            bnd = self.mol.GetBondById(i)
            # bnd = mol.GetBond(i)
            if bnd is None:
                continue
            bond_idx.append([bnd.GetBeginAtomIdx() - 1, bnd.GetEndAtomIdx() - 1])
            if not self._make_directed:
                # Add a bond with opposite direction but same properties
                bond_idx.append([bnd.GetEndAtomIdx() - 1, bnd.GetBeginAtomIdx() - 1])
        # Sort bond indices
        bond_idx, _ = self._sort_bonds(bond_idx)
        return bond_idx

    @property
    def edge_number(self):
        """Return a list of edge number that represents the bond order."""
        bond_number = []
        bond_idx = []
        for i in range(self.mol.NumBonds()):
            bnd = self.mol.GetBondById(i)
            # bnd = mol.GetBond(i)
            if bnd is None:
                continue
            bond_idx.append([bnd.GetBeginAtomIdx() - 1, bnd.GetEndAtomIdx() - 1])
            bond_number.append(bnd.GetBondOrder())
            if not self._make_directed:
                # Add a bond with opposite direction but same properties
                bond_idx.append([bnd.GetEndAtomIdx() - 1, bnd.GetBeginAtomIdx() - 1])
                bond_number.append(bnd.GetBondOrder())
                # Sort bond indices
        bond_idx, bond_number = self._sort_bonds(bond_idx, bond_number)
        return bond_idx, bond_number

    def edge_attributes(self, properties: list, encoder: dict):
        """Make edge attributes.

        Args:
            properties (list): List of string identifier for a molecular property. Must match backend features.
            encoder (dict): A dictionary of callable encoder function or class for each string identifier.

        Returns:
            list: List of attributes after processed by the encoder.
        """
        m = self.mol
        edges = self._check_properties_list(properties, sorted(self.bond_fun_dict.keys()), "Bond")
        encoder = self._check_encoder(encoder, sorted(self.bond_fun_dict.keys()))

        # Collect info about bonds
        bond_info = []
        bond_idx = []
        for i in range(m.NumBonds()):
            x = m.GetBondById(i)
            attr = []
            for k in edges:
                if isinstance(k, str):
                    temp = encoder[k](self.bond_fun_dict[k](x)) if k in encoder else self.bond_fun_dict[k](x)
                else:
                    temp = k(x)
                if isinstance(temp, np.ndarray):
                    temp = temp.tolist()
                if isinstance(temp, (list, tuple)):
                    attr += list(temp)
                else:
                    attr.append(temp)
            bond_info.append(attr)
            bond_idx.append([x.GetBeginAtomIdx() - 1, x.GetEndAtomIdx() - 1])
            # Add a bond with opposite direction but same properties
            if not self._make_directed:
                bond_info.append(attr)
                bond_idx.append([x.GetEndAtomIdx() - 1, x.GetBeginAtomIdx() - 1])

        # Sort directed bonds
        bond_idx, bond_info = self._sort_bonds(bond_idx, bond_info)
        return bond_idx, bond_info

    def node_attributes(self, properties: list, encoder: dict):
        """Make node attributes.

        Args:
            properties (list): List of string identifier for a molecular property. Must match backend features.
            encoder (dict): A dictionary of callable encoder function or class for each string identifier.

        Returns:
            list: List of attributes after processed by the encoder.
        """
        m = self.mol
        nodes = self._check_properties_list(properties, sorted(self.atom_fun_dict.keys()), "Atom")
        encoder = self._check_encoder(encoder, sorted(self.atom_fun_dict.keys()))
        # Collect info about atoms
        atom_info = []
        for i in range(m.NumAtoms()):
            atm = m.GetAtomById(i)
            attr = []
            for k in nodes:
                if isinstance(k, str):
                    temp = encoder[k](self.atom_fun_dict[k](atm)) if k in encoder else self.atom_fun_dict[k](atm)
                else:
                    temp = k(atm)
                if isinstance(temp, np.ndarray):
                    temp = temp.tolist()
                if isinstance(temp, (list, tuple)):
                    attr += list(temp)
                else:
                    attr.append(temp)
            atom_info.append(attr)
        return atom_info

    def graph_attributes(self, properties: list, encoder: dict):
        """Make graph attributes.

        Args:
            properties (list): List of string identifier for a molecular property. Must match backend features.
            encoder (dict): A dictionary of callable encoder function or class for each string identifier.

        Returns:
            list: List of attributes after processed by the encoder.
        """
        graph = self._check_properties_list(properties, sorted(self.mol_fun_dict.keys()), "Molecule")
        m = self.mol
        encoder = self._check_encoder(encoder, sorted(self.mol_fun_dict.keys()))
        # Mol info
        attr = []
        for k in graph:
            if isinstance(k, str):
                temp = encoder[k](self.mol_fun_dict[k](m)) if k in encoder else self.mol_fun_dict[k](m)
            else:
                temp = k(m)
            if isinstance(temp, np.ndarray):
                temp = temp.tolist()
            if isinstance(temp, (list, tuple)):
                attr += list(temp)
            else:
                attr.append(temp)
        return attr



import numpy as np
from kgcnn.utils.serial import serialize, deserialize
from kgcnn.graph.base import GraphPostProcessorBase, GraphDict


class ExtensiveEnergyForceScalerPostprocessor(GraphPostProcessorBase):
    r"""Postprocessor to inverse-transform energies and forces from a graph dictionary.
    Note that this postprocessor requires the input- or pre-graph in call method as well.

    Args:
        scaler: Fitted instance of :obj:`ExtensiveEnergyForceScaler` .
        energy (str): Name of energy property in :obj:`graph` dict. Default is 'energy'.
        force (str): Name of force property in :obj:`graph` dict. Default is 'forces'.
        atomic_number (str): Name of atomic_number property in :obj:`pre_graph` dict. Default is 'node_number'.
            Must be given in additional :obj:`pre_graph` that is passed to call.
    """

    def __init__(self, scaler, energy: str = "energy",
                 force: str = "forces", atomic_number: str = "node_number",
                 name="extensive_energy_force_scaler", **kwargs):
        super(ExtensiveEnergyForceScalerPostprocessor, self).__init__(name=name, **kwargs)
        if isinstance(scaler, dict):
            self.scaler = deserialize(scaler)
        else:
            self.scaler = scaler
        self._to_obtain_pre = {"atomic_number": atomic_number}
        self._to_obtain = {"y": energy, "force": force}
        self._to_assign = [energy, force]
        self._config_kwargs.update(
            {"scaler": serialize(self.scaler), "energy": energy, "force": force, "atomic_number": atomic_number})

    def call(self, y, force, atomic_number):
        # Need batch with one energy etc.
        # Can copy data.
        energy, forces = self.scaler.inverse_transform(
            X=None, y=np.array([y]), force=[force], atomic_number=[atomic_number])
        return energy[0], forces[0]



import numpy as np
from kgcnn.data.datasets.MoleculeNetDataset2018 import MoleculeNetDataset2018


class ClinToxDataset(MoleculeNetDataset2018):
    r"""Store and process 'ClinTox' dataset from `MoleculeNet <https://moleculenet.org/>`_ database.
    Class inherits from :obj:`MoleculeNetDataset2018` and downloads dataset on class initialization.

    Compare reference: `DeepChem <https://deepchem.readthedocs.io/en/latest/api_reference/
    moleculenet.html#clintox-datasets>`_ reading:
    Qualitative data of drugs approved by the FDA and those that have failed clinical trials for toxicity reasons.
    Random splitting is recommended for this dataset.

    The ClinTox dataset compares drugs approved by the FDA and drugs that have failed clinical trials for toxicity.
    The dataset includes two classification tasks for 1491 drug compounds with known chemical structures:

        - clinical trial toxicity (or absence of toxicity)
        - FDA approval status.

    List of FDA-approved drugs are compiled from the SWEETLEAD database, and list of drugs that failed clinical trials
    for toxicity reasons are compiled from the Aggregate Analysis of ClinicalTrials.gov(AACT) database.

    References:

        (1) Gayvert, Kaitlyn M., Neel S. Madhukar, and Olivier Elemento.
            A data-driven approach to predicting successes and failures of clinica trials.
            Cell chemical biology 23.10 (2016): 1294-1301.
        (2) Artemov, Artem V., et al. Integrated deep learned transcriptomic and
            structure-based predictor of clinical trials outcomes. bioRxiv (2016): 095653.
        (3) Novick, Paul A., et al. SWEETLEAD: an in silico database of approved drugs,
            regulated chemicals, and herbal isolates for computer-aided drug discovery.
            PloS one 8.11 (2013): e79568.
        (4) Aggregate Analysis of ClincalTrials.gov (AACT) Database.
            `<https://www.ctti-clinicaltrials.org/aact-database>`_ .

    """

    def __init__(self, reload=False, verbose: int = 10, label_index: int = 0):
        """Initialize ClinTox dataset.

        Args:
            reload (bool): Whether to reload the data and make new dataset. Default is False.
            verbose (int): Print progress or info for processing where 60=silent. Default is 10.
            label_index (int): Which information should be taken as binary label. Default is 0.
                Determines the positive class label, which is 'approved' by default.
        """
        self.label_index = label_index
        super(ClinToxDataset, self).__init__("ClinTox", reload=reload, verbose=verbose)

    def read_in_memory(self, **kwargs):
        super(ClinToxDataset, self).read_in_memory(**kwargs)

        # Strictly one class so we pick drug approved for positive label.
        graph_labels = self.obtain_property("graph_labels")
        graph_labels = [np.array([x[self.label_index]], dtype="float") if x is not None else None for x in graph_labels]
        self.assign_property("graph_labels", graph_labels)



import os
import unittest

from visual_graph_datasets.testing import IsolatedConfig

from kgcnn.data.visual_graph import VisualGraphDataset


class TestVisualGraphDataset(unittest.TestCase):

    def test_basically_works(self):
        """
        If the VisualGraphDataset class generally works. Attempts to load the "mock" dataset into memory.
        """
        with IsolatedConfig() as config:
            # The dataset "mock" is a small dataset which should always be available for testing purposes.
            # the dataset contains exactly 100 randomly generated graphs, which is what we test for later
            # on as well to confirm that it works.
            vgd = VisualGraphDataset('mock', config=config)

            vgd.ensure()
            self.assertTrue(os.path.exists(vgd.data_directory))

            vgd.read_in_memory()
            self.assertNotEqual(0, len(vgd))
            self.assertEqual(100, len(vgd))

    def test_has_importances_works(self):
        """
        If the method "has_importances" works as expected
        """
        with IsolatedConfig() as config:
            vgd = VisualGraphDataset('mock', config=config)
            vgd.ensure()

            vgd.read_in_memory()
            # For this dataset we know that it does in fact contain ground truth explanation annotations
            # for 2 channels - one for each of the two classes.
            self.assertTrue(vgd.has_importances('2'))



import tensorflow as tf
import numpy as np
from kgcnn.layers.base import GraphBaseLayer
from kgcnn.layers.gather import GatherNodesOutgoing, GatherNodesSelection, GatherNodesIngoing
from kgcnn.layers.geom import NodeDistanceEuclidean, NodePosition
from kgcnn.layers.pooling import PoolingLocalEdges
from kgcnn.layers.modules import LazyMultiply, LazySubtract, ExpandDims

ks = tf.keras

# Parameters for Weighted Atom-Centered Symmetry Functions (wACSF).
# We set the unoptimized C default values as default for all 118 atomic species.
# They are taken from https://aip.scitation.org/doi/suppl/10.1063/1.5019667/suppl_file/si.pdf .
# We chose the 22:10 radial:angular as default values. Other combinations ranging from 0:32 to 32:0 .
radial_eta_mu_defaults = np.array([[
    [4.5000000, 7.5000000, 8.0], [4.5000000, 7.1667000, 8.0], [4.5000000, 6.8333000, 8.0], [4.5000000, 6.5000000, 8.0],
    [4.5000000, 6.1667000, 8.0], [4.5000000, 5.8333000, 8.0], [4.5000000, 5.5000000, 8.0], [4.5000000, 5.1667000, 8.0],
    [4.5000000, 4.8333000, 8.0], [4.5000000, 4.5000000, 8.0], [4.5000000, 4.1667000, 8.0], [4.5000000, 3.8333000, 8.0],
    [4.5000000, 3.5000000, 8.0], [4.5000000, 3.1667000, 8.0], [4.5000000, 2.8333000, 8.0], [4.5000000, 2.5000000, 8.0],
    [4.5000000, 2.1667000, 8.0], [4.5000000, 1.8333000, 8.0], [4.5000000, 1.5000000, 8.0], [4.5000000, 1.1667000, 8.0],
    [4.5000000, 0.8333000, 8.0], [4.5000000, 0.5000000, 8.0]
]] * 118)
angular_eta_mu_lambda_zeta_defaults = np.array(
    [[[0.03306120, 0.0, -1.0, 1.0, 8.00], [0.03306120, 0.0, 1.0, 1.0, 8.00], [0.04986150, 0.0, -1.0, 1.0, 8.00],
      [0.04986150, 0.0, 1.0, 1.0, 8.00], [0.08367770, 0.0, -1.0, 1.0, 8.00], [0.08367770, 0.0, 1.0, 1.0, 8.00],
      [0.16857440, 0.0, -1.0, 1.0, 8.00], [0.16857440, 0.0, 1.0, 1.0, 8.00], [0.50000000, 0.0, -1.0, 1.0, 8.00],
      [0.50000000, 0.0, 1.0, 1.0, 8.00]]] * 118)
# Update optimized parameters for H, C, F, O, N .
radial_eta_mu_defaults[1] = np.array(
    [[4.3814240, 2.4432533, 8.0], [4.4149216, 6.1667000, 8.0], [4.4493159, 3.8332928, 8.0], [4.4815889, 3.1667000, 8.0],
     [4.4865336, 3.5000000, 8.0], [4.4937252, 4.8332601, 8.0], [4.4986009, 5.5010689, 8.0], [4.4997975, 2.1558150, 8.0],
     [4.4999291, 7.4956238, 8.0], [4.4999905, 6.4964721, 8.0], [4.5000000, 5.2730567, 8.0], [4.5000011, 1.1606611, 8.0],
     [4.5000039, 0.4918029, 8.0], [4.5000398, 0.8333002, 8.0], [4.5002241, 1.8334858, 8.0], [4.5036121, 2.9532907, 8.0],
     [4.5059680, 5.8330954, 8.0], [4.5062245, 6.8333000, 8.0], [4.5064492, 7.1666408, 8.0], [4.5675834, 4.5046412, 8.0],
     [4.5899630, 4.1663934, 8.0], [4.6604619, 1.4989046, 8.0]])
angular_eta_mu_lambda_zeta_defaults[1] = np.array(
    [[0.01919876, 0.0, -1.0, 1.0, 8.00], [0.03306120, 0.0, 1.0, 1.0, 8.00], [0.03447591, 0.0, -1.0, 2.0, 8.00],
     [0.04986150, 0.0, -1.0, 1.0, 8.00], [0.08235996, 0.0, -1.0, 1.0, 8.00], [0.08367770, 0.0, 1.0, 2.0, 8.00],
     [0.16857440, 0.0, -1.0, 2.0, 8.00], [0.16953019, 0.0, 1.0, 2.0, 8.00], [0.50000000, 0.0, -1.0, 1.0, 8.00],
     [0.50000000, 0.0, 1.0, 1.0, 8.00]]
)
radial_eta_mu_defaults[6] = np.array(
    [[4.4534562, 3.8338779, 8.0], [4.4804536, 5.1667000, 8.0], [4.4918531, 4.8333000, 8.0], [4.4991551, 0.8333000, 8.0],
     [4.4998245, 2.4997843, 8.0], [4.4999043, 1.5993354, 8.0], [4.4999230, 5.8324479, 8.0], [4.4999997, 6.5117668, 8.0],
     [4.5000000, 6.1860418, 8.0], [4.5000000, 3.1584848, 8.0], [4.5000001, 4.4991770, 8.0], [4.5000132, 6.8333000, 8.0],
     [4.5001099, 7.4999458, 8.0], [4.5002911, 2.1638163, 8.0], [4.5004121, 0.3038147, 8.0], [4.5014183, 5.4361520, 8.0],
     [4.5023686, 1.8302079, 8.0], [4.5027053, 3.5002020, 8.0], [4.5074706, 7.2656992, 8.0], [4.5104731, 1.1645345, 8.0],
     [4.5515436, 4.1666210, 8.0], [4.5906934, 2.8328539, 8.0]])
angular_eta_mu_lambda_zeta_defaults[6] = np.array(
    [[0.03392104, 0.0, 1.0, 2.0, 8.00], [0.04232205, 0.0, -1.0, 2.0, 8.00], [0.04663133, 0.0, -1.0, 2.0, 8.00],
     [0.05097753, 0.0, -1.0, 1.0, 8.00], [0.08367770, 0.0, -1.0, 3.0, 8.00], [0.08368552, 0.0, -1.0, 1.0, 8.00],
     [0.16857440, 0.0, 1.0, 2.0, 8.00], [0.25965214, 0.0, 1.0, 2.0, 8.00], [0.47868498, 0.0, -1.0, 1.0, 8.00],
     [0.56008922, 0.0, 1.0, 1.0, 8.00]]
)
radial_eta_mu_defaults[9] = np.array(
    [[4.3755129, 4.3891938, 8.0], [4.4739340, 6.5000000, 8.0], [4.4934064, 1.1667000, 8.0], [4.4955168, 3.1421155, 8.0],
     [4.4982145, 6.1665305, 8.0], [4.4986470, 0.8333000, 8.0], [4.4989965, 7.4999930, 8.0], [4.4999944, 4.1666929, 8.0],
     [4.4999998, 6.8333000, 8.0], [4.4999998, 2.8332546, 8.0], [4.4999999, 5.8424471, 8.0], [4.5000000, 5.4954750, 8.0],
     [4.5000000, 4.8332314, 8.0], [4.5000000, 2.1675890, 8.0], [4.5000000, 1.8332281, 8.0], [4.5000016, 3.8334246, 8.0],
     [4.5000047, 7.1667013, 8.0], [4.5002559, 0.5866000, 8.0], [4.5004160, 5.1667000, 8.0], [4.5165951, 1.5000000, 8.0],
     [4.5200686, 3.5000000, 8.0], [4.5333816, 2.5065631, 8.0]])
angular_eta_mu_lambda_zeta_defaults[9] = np.array(
    [[0.03302291, 0.0, -1.0, 1.0, 8.00], [0.03306120, 0.0, 1.0, 1.0, 8.00], [0.04986182, 0.0, -1.0, 1.0, 8.00],
     [0.05024516, 0.0, 1.0, 1.0, 8.00], [0.08367770, 0.0, -1.0, 2.0, 8.00], [0.08367770, 0.0, 1.0, 1.0, 8.00],
     [0.16857440, 0.0, 1.0, 2.0, 8.00], [0.16857440, 0.0, 1.0, 1.0, 8.00], [0.49854417, 0.0, 1.0, 1.0, 8.00],
     [0.50003188, 0.0, 1.0, 1.0, 8.00]]
)
radial_eta_mu_defaults[8] = np.array(
    [[4.3810310, 6.1634839, 8.0], [4.4613793, 4.7025864, 8.0], [4.4708866, 0.8337341, 8.0], [4.4741510, 7.1654091, 8.0],
     [4.4934208, 0.4909963, 8.0], [4.4999979, 6.7895512, 8.0], [4.4999984, 2.5548726, 8.0], [4.4999996, 5.1188661, 8.0],
     [4.4999999, 6.4941667, 8.0], [4.5000000, 7.5269793, 8.0], [4.5000000, 3.5008886, 8.0], [4.5000000, 3.1673548, 8.0],
     [4.5000000, 3.0664616, 8.0], [4.5000000, 1.1771551, 8.0], [4.5000010, 5.9628537, 8.0], [4.5000305, 1.6101757, 8.0],
     [4.5000505, 3.8352953, 8.0], [4.5000877, 5.2259600, 8.0], [4.5003116, 2.1667069, 8.0], [4.5007395, 1.8264699, 8.0],
     [4.5037122, 4.1666850, 8.0], [4.5108571, 4.7474998, 8.0]])
angular_eta_mu_lambda_zeta_defaults[8] = np.array(
    [[0.03306120, 0.0, -1.0, 1.0, 8.00], [0.04970837, 0.0, -1.0, 1.0, 8.00], [0.04986150, 0.0, 1.0, 1.0, 8.00],
     [0.05680303, 0.0, -1.0, 1.0, 8.00], [0.08366889, 0.0, 1.0, 1.0, 8.00], [0.08367770, 0.0, -1.0, 1.0, 8.00],
     [0.16857440, 0.0, 1.0, 2.0, 8.00], [0.16857440, 0.0, 1.0, 1.0, 8.00], [0.50000000, 0.0, -1.0, 1.0, 8.00],
     [0.50000000, 0.0, 1.0, 1.0, 8.00]]
)
radial_eta_mu_defaults[7] = np.array(
    [[4.4148209, 6.8472192, 8.0], [4.4481457, 1.0317668, 8.0], [4.4608631, 0.4581865, 8.0], [4.4797608, 3.9799903, 8.0],
     [4.4951823, 0.8337448, 8.0], [4.4987030, 7.4573738, 8.0], [4.4998898, 6.9779219, 8.0], [4.4999884, 2.1676806, 8.0],
     [4.4999920, 3.1667000, 8.0], [4.5000000, 6.1662980, 8.0], [4.5000000, 5.8328781, 8.0], [4.5000000, 5.4757644, 8.0],
     [4.5000000, 5.1666981, 8.0], [4.5000000, 3.3518840, 8.0], [4.5000000, 2.4982784, 8.0], [4.5001975, 4.8204812, 8.0],
     [4.5004709, 6.5050056, 8.0], [4.5016337, 4.1667000, 8.0], [4.5077984, 1.4994658, 8.0], [4.5228543, 2.8333000, 8.0],
     [4.5780852, 4.5000000, 8.0], [4.5897686, 1.8340302, 8.0]])
angular_eta_mu_lambda_zeta_defaults[7] = np.array(
    [[0.03306120, 0.0, -1.0, 1.0, 8.00], [0.03306120, 0.0, 1.0, 1.0, 8.00], [0.04907151, 0.0, -1.0, 2.0, 8.00],
     [0.07102913, 0.0, 1.0, 1.0, 8.00], [0.07116163, 0.0, 1.0, 1.0, 8.00], [0.08384128, 0.0, -1.0, 1.0, 8.00],
     [0.16857391, 0.0, -1.0, 2.0, 8.00], [0.16902220, 0.0, 1.0, 1.0, 8.00], [0.50000000, 0.0, -1.0, 1.0, 8.00],
     [0.50000000, 0.0, 1.0, 1.0, 8.00]]
)


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='wACSFRad')
class wACSFRad(GraphBaseLayer):
    r"""Weighted atom-centered symmetry functions (wACSF) for high-dimensional neural network potentials (HDNNPs).
    From `Gastegger et al. (2017) <https://arxiv.org/abs/1712.05861>`_ .
    Default values can be found in `<https://aip.scitation.org/doi/suppl/10.1063/1.5019667>`_ .
    This layer implements the radial part :math:`W_{i}^{rad}` :

    .. math::

        W_{i}^{rad} = \sum_{j \neq i} \; g(Z_j) \; e^{−\eta \, (r_{ij} − \mu)^{2} } \; f_{ij}

    Here, for each atom type there is a set of parameters :math:`\eta` and :math:`\mu` .
    The cutoff function :math:`f_ij = f_c(r_{ij})` is given by:

    .. math::

        f_c(r_{ij}) = 0.5 [\cos{\frac{\pi r_{ij}}{R_c}} + 1]

    The cutoff radius is implemented as a float and can not be changed dependent on the atom type.
    Not that the parameters :math:`\eta` and :math:`\mu` can be made trainable.

    """

    def __init__(self, eta_mu: list = None, cutoff: float = 8.0, add_eps: bool = False,
                 use_external_weights: bool = False, param_constraint=None, param_regularizer=None,
                 param_initializer="zeros", **kwargs):
        super(wACSFRad, self).__init__(**kwargs)
        self.cutoff = cutoff
        self.add_eps = add_eps
        if eta_mu is None:
            eta_mu = radial_eta_mu_defaults[:, :, :2]
        self.eta_mu = np.array(eta_mu, dtype="float").tolist()
        self.use_external_weights = use_external_weights
        self.lazy_mult = LazyMultiply()
        self.layer_pos = NodePosition()
        self.layer_gather_out = GatherNodesOutgoing()
        self.layer_gather_in = GatherNodesIngoing()
        self.layer_exp_dims = ExpandDims(axis=2)
        self.layer_dist = NodeDistanceEuclidean(add_eps=add_eps)
        self.pool_sum = PoolingLocalEdges(pooling_method="sum")

        # We can do this in init since weights do not depend on input shape.
        self.param_initializer = ks.initializers.deserialize(param_initializer)
        self.param_regularizer = ks.regularizers.deserialize(param_regularizer)
        self.param_constraint = ks.constraints.deserialize(param_constraint)
        np_eta_mu = np.array(eta_mu)
        weight_shape = np_eta_mu.shape
        self._weight_eta_mu = self.add_weight(
            "eta_mu",
            shape=weight_shape,
            initializer=self.param_initializer,
            regularizer=self.param_regularizer,
            constraint=self.param_constraint,
            dtype=self.dtype, trainable=False
        )
        self.set_weights([np_eta_mu])

    @staticmethod
    def _compute_fc(inputs: tf.Tensor, cutoff: float):
        fc = tf.clip_by_value(inputs, -cutoff, cutoff)
        fc = (tf.math.cos(fc * np.pi / cutoff) + 1.0) * 0.5
        # fc = tf.where(tf.abs(inputs) < self.cutoff, fc, tf.zeros_like(fc))
        return fc

    def _compute_gaussian_expansion(self, inputs: tf.Tensor):
        rij, zi = inputs
        params = tf.gather(self._weight_eta_mu, zi, axis=0)
        # eta, mu = tf.split(params, [1, 1], axis=-1)
        # eta, mu = tf.squeeze(eta, axis=-1), tf.squeeze(mu, axis=-1)
        eta, mu = tf.gather(params, 0, axis=-1), tf.gather(params, 1, axis=-1)
        arg = tf.square(rij - mu) * eta
        return tf.exp(-arg)

    def build(self, input_shape):
        super(wACSFRad, self).build(input_shape)

    def call(self, inputs, mask=None, **kwargs):
        r"""Forward pass.

        Args:
            inputs: [z, xyz, ij] or [z, xyz, ij, w]  if `use_external_weights`

                - z (tf.RaggedTensor): Atomic numbers of shape (batch, [N])
                - xyz (tf.RaggedTensor): Node coordinates of shape (batch, [N], 3)
                - ij (tf.RaggedTensor): Edge indices referring to nodes of shape (batch, [M], 2)
                - w (tf.RaggedTensor): Edge weight tensor of shape (batch, [M], 1)

            mask: Boolean mask for inputs. Not used. Defaults to None.

        Returns:
            tf.RaggedTensor: Atomic representation of shape `(batch, None, units)` .
        """
        if self.use_external_weights:
            z, xyz, eij, w = self.assert_ragged_input_rank(inputs, mask=mask, ragged_rank=1)
            z = self.map_values(tf.cast, z, dtype=eij.dtype)
        else:
            z, xyz, eij = self.assert_ragged_input_rank(inputs, mask=mask, ragged_rank=1)
            z = self.map_values(tf.cast, z, dtype=eij.dtype)
            zj = self.layer_gather_out([z, eij], **kwargs)
            w = self.layer_exp_dims(zj, **kwargs)
        w = self.map_values(tf.cast, w, dtype=self.dtype)
        xi, xj = self.layer_pos([xyz, eij], **kwargs)
        rij = self.layer_dist([xi, xj], **kwargs)
        fc = self.map_values(self._compute_fc, rij, cutoff=self.cutoff)
        zi = self.layer_gather_in([z, eij], **kwargs)
        gij = self.map_values(self._compute_gaussian_expansion, [rij, zi])
        rep = self.lazy_mult([gij, fc, w], **kwargs)
        return self.pool_sum([xyz, rep, eij], **kwargs)

    def get_config(self):
        config = super(wACSFRad, self).get_config()
        config.update({"eta_mu": self.eta_mu, "cutoff": self.cutoff, "use_external_weights": self.use_external_weights,
                       "add_eps": self.add_eps,
                       "param_constraint": ks.constraints.serialize(self.param_constraint),
                       "param_regularizer": ks.regularizers.serialize(self.param_regularizer),
                       "param_initializer": ks.initializers.serialize(self.param_initializer)
                       })
        return config


@tf.keras.utils.register_keras_serializable(package='kgcnn', name='wACSFAng')
class wACSFAng(GraphBaseLayer):
    r"""Weighted atom-centered symmetry functions (wACSF) for high-dimensional neural network potentials (HDNNPs).
    From `Gastegger et al. (2017) <https://arxiv.org/abs/1712.05861>`_ .
    Default values can be found in `<https://aip.scitation.org/doi/suppl/10.1063/1.5019667>`_ .
    This layer implements the angular part :math:`W_{i}^{ang}` :

    .. math::

        W_{i}^{ang} = 2^{1−\zeta} \; \sum_{j\neq i} \sum_{k\neq j,i} \; h(Z_j , Z_k) \;
        (1 + \lambda\, \cos{\theta_{ijk}})^\zeta\;
        \times \; e^{−\eta (r_{ij} −\mu )^2} \; e^{−\eta (r_{ik} −\mu )^2} \; e^{−\eta (r_{jk} −\mu )^2} \;
        \times \; f_{ij} \; f_{ik} \; f_{jk}

    Here, for each atom type there is a set of parameters :math:`\eta` , :math:`\mu` , :math:`\lambda`
    and :math:`\zeta`.
    The cutoff function :math:`f_ij = f_c(r_{ij})` is given by:

    .. math::

        f_c(r_{ij}) = 0.5 [\cos{\frac{\pi r_{ij}}{R_c}} + 1]

    The cutoff radius is implemented as a float and can not be changed dependent on the atom type.
    Not that the parameters :math:`\eta` to :math:`\lambda` can be made trainable.

    """
    def __init__(self, eta_mu_lambda_zeta: list = None, cutoff: float = 8.0,
                 add_eps: bool = False, use_external_weights: bool = False, param_initializer="zeros",
                 param_regularizer=None, param_constraint=None, **kwargs):
        super(wACSFAng, self).__init__(**kwargs)
        self.add_eps = add_eps
        self.cutoff = cutoff
        if eta_mu_lambda_zeta is None:
            eta_mu_lambda_zeta = angular_eta_mu_lambda_zeta_defaults[:, :, :4]
        self.eta_mu_lambda_zeta = np.array(eta_mu_lambda_zeta, dtype="float").tolist()
        self.use_external_weights = use_external_weights
        self.lazy_mult = LazyMultiply()
        self.layer_pos = NodePosition(selection_index=[0, 1, 2])
        self.layer_dist = NodeDistanceEuclidean(add_eps=add_eps)
        self.pool_sum = PoolingLocalEdges(pooling_method="sum")
        self.lazy_sub = LazySubtract()
        self.layer_gather_in = GatherNodesIngoing()
        self.layer_gather_1 = GatherNodesSelection(selection_index=1)
        self.layer_gather_2 = GatherNodesSelection(selection_index=2)
        self.layer_exp_dims = ExpandDims(axis=2)
        # We can do this in init since weights do not depend on input shape.
        self.param_initializer = ks.initializers.deserialize(param_initializer)
        self.param_regularizer = ks.regularizers.deserialize(param_regularizer)
        self.param_constraint = ks.constraints.deserialize(param_constraint)
        np_eta_mu_lambda_zeta = np.array(eta_mu_lambda_zeta)
        weight_shape = np_eta_mu_lambda_zeta.shape
        self._weight_eta_mu_lambda_zeta = self.add_weight(
            "eta_mu_lambda_zeta",
            shape=weight_shape,
            initializer=self.param_initializer,
            regularizer=self.param_regularizer,
            constraint=self.param_constraint,
            dtype=self.dtype, trainable=False
        )
        self.set_weights([np_eta_mu_lambda_zeta])

    @staticmethod
    def _compute_fc(inputs: tf.Tensor, cutoff: float):
        fc = tf.clip_by_value(inputs, -cutoff, cutoff)
        fc = (tf.math.cos(fc * np.pi / cutoff) + 1.0) * 0.5
        # fc = tf.where(tf.abs(inputs) < self.cutoff, fc, tf.zeros_like(fc))
        return fc

    def _compute_gaussian_expansion(self, inputs: tf.Tensor):
        rij, zi = inputs
        params = tf.gather(self._weight_eta_mu_lambda_zeta, zi, axis=0)
        eta, mu = tf.gather(params, 0, axis=-1), tf.gather(params, 1, axis=-1)
        arg = tf.square(rij - mu) * eta
        return tf.exp(-arg)

    def _compute_pow_cos_angle_(self, inputs: list):
        vij, vik, rij, rik, zi = inputs
        params = tf.gather(self._weight_eta_mu_lambda_zeta, zi, axis=0)
        lamda, zeta = tf.gather(params, 2, axis=-1), tf.gather(params, 3, axis=-1)
        cos_theta = tf.reduce_sum(vij * vik, axis=-1, keepdims=True) / rij / rik
        cos_term = cos_theta * lamda + 1.0
        cos_term = tf.pow(cos_term, zeta)
        return cos_term

    def _compute_pow_scale(self, inputs: tf.Tensor):
        to_scale, zi = inputs
        params = tf.gather(self._weight_eta_mu_lambda_zeta, zi, axis=0)
        zeta = tf.gather(params, 3, axis=-1)
        scale = tf.ones_like(to_scale) * 2.0
        scaled = tf.pow(scale, 1.0 - zeta) * to_scale
        return scaled

    def build(self, input_shape):
        super(wACSFAng, self).build(input_shape)

    def call(self, inputs, mask=None, **kwargs):
        r"""Forward pass.

        Args:
            inputs (list): [z, xyz, ijk] or [z, xyz, ijk, w]  if `use_external_weights`

                - z (tf.RaggedTensor): Atomic numbers of shape (batch, [N])
                - xyz (tf.RaggedTensor): Node coordinates of shape (batch, [N], 3)
                - ijk (tf.RaggedTensor): Angle indices referring to nodes of shape (batch, [M], 3)
                - w (tf.RaggedTensor): Angle weight tensor of shape (batch, [M], 1)

            mask (list): Boolean mask for inputs. Not used. Defaults to None.

        Returns:
            tf.RaggedTensor: Atomic representation of shape `(batch, None, units)` .
        """
        if self.use_external_weights:
            z, xyz, ijk, w = self.assert_ragged_input_rank(inputs, mask=mask, ragged_rank=1)
            z = self.map_values(tf.cast, z, dtype=ijk.dtype)
            w = self.map_values(tf.cast, w, dtype=self.dtype)
        else:
            z, xyz, ijk = self.assert_ragged_input_rank(inputs, mask=mask, ragged_rank=1)
            z = self.map_values(tf.cast, z, dtype=ijk.dtype)
            w1 = self.layer_gather_1([z, ijk], **kwargs)[0]
            w2 = self.layer_gather_2([z, ijk], **kwargs)[0]
            w1 = self.map_values(tf.cast, w1, dtype=self.dtype)
            w2 = self.map_values(tf.cast, w2, dtype=self.dtype)
            w = self.lazy_mult([w1, w2], **kwargs)
            w = self.layer_exp_dims(w, **kwargs)
        xi, xj, xk = self.layer_pos([xyz, ijk])
        rij = self.layer_dist([xi, xj], **kwargs)
        rik = self.layer_dist([xi, xk], **kwargs)
        rjk = self.layer_dist([xj, xk], **kwargs)
        fij = self.map_values(self._compute_fc, rij, cutoff=self.cutoff)
        fik = self.map_values(self._compute_fc, rik, cutoff=self.cutoff)
        fjk = self.map_values(self._compute_fc, rjk, cutoff=self.cutoff)
        zi = self.layer_gather_in([z, ijk], **kwargs)
        gij = self.map_values(self._compute_gaussian_expansion, [rij, zi])
        gik = self.map_values(self._compute_gaussian_expansion, [rik, zi])
        gjk = self.map_values(self._compute_gaussian_expansion, [rjk, zi])
        vij = self.lazy_sub([xi, xj], **kwargs)
        vik = self.lazy_sub([xi, xk], **kwargs)
        pow_cos_theta = self.map_values(self._compute_pow_cos_angle_, [vij, vik, rij, rik, zi])
        rep = self.lazy_mult([pow_cos_theta, gij, gik, gjk, fij, fik, fjk, w], **kwargs)
        pool_ang = self.pool_sum([xyz, rep, ijk], **kwargs)
        return self.map_values(self._compute_pow_scale, [pool_ang, z])

    def get_config(self):
        config = super(wACSFAng, self).get_config()
        config.update({"eta_mu_lambda_zeta": self.eta_mu_lambda_zeta, "cutoff": self.cutoff,
                       "use_external_weights": self.use_external_weights,
                       "add_eps": self.add_eps,
                       "param_constraint": ks.constraints.serialize(self.param_constraint),
                       "param_regularizer": ks.regularizers.serialize(self.param_regularizer),
                       "param_initializer": ks.initializers.serialize(self.param_initializer)
                       })
        return config
